#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for making plots for regression models (e.g. energy, dir, vtx, bjorkeny, ...).
"""

import warnings
import math
import numpy as np
import matplotlib as mpl
from matplotlib import pyplot as plt

from orcanet_contrib.plotting.utils import correct_reco_energy, select_ic, get_event_selection_mask


# --------------------------- Code for 2d plots --------------------------- #


def make_2d_prop_to_prop_plot(pred_file, prop_1_name, prop_2_name, savefolder, savename, reco_energy_correction=None,
                              cuts=None, title_prefix='OrcaNet: '):
    """
    Function that makes a 2d plot of property 1 vs property 2.

    The available properties are grouped in the function get_make_2d_energy_resolution_plot_properties_dict().

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the regression predictions of a nn model.
    prop_1_name : str
        Name of the first property that should be plotted on the X-axis.
    prop_2_name : str
        Name of the second property that should be plotted on the Y-axis.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    savename : str
        Savename with which the plots file should be saved.
    reco_energy_correction : None/str
        If on of the props is the reco energy, specifies if the energy should be corrected or not.
        If not None, the string specifies the metric that should be used for the correction.
        Available: 'median' and 'mean'.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.
    title_prefix : str
        String that is used as a prefix for the title of the plot.

    """
    properties = get_make_2d_energy_resolution_plot_properties_dict()
    if reco_energy_correction is True:
        properties['energy_reco']['ax_label'] = 'Corrected reconstructed energy (GeV)'  # add corrected info in str

    mc_info = pred_file['mc_info']
    dset_key_prop_1, dset_key_prop_2 = properties[prop_1_name]['dset_key'], properties[prop_2_name]['dset_key']
    prop_dset_1, prop_dset_2 = pred_file[dset_key_prop_1], pred_file[dset_key_prop_2]  # e.g. y_pred dset

    if cuts is not None:
        assert isinstance(cuts, str)
        evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
        mc_info = mc_info[evt_sel_mask]
        prop_dset_1 = prop_dset_1[evt_sel_mask]
        prop_dset_2 = prop_dset_2[evt_sel_mask]

    prop_1 = get_property_info_to_plot(mc_info, prop_dset_1, properties, prop_1_name, reco_energy_correction)
    prop_2 = get_property_info_to_plot(mc_info, prop_dset_2, properties, prop_2_name, reco_energy_correction)

    ic_list = {'muon-CC': {'title': 'Track like (' + r'$\nu_{\mu}-CC$)'},
               'elec-CC': {'title': 'Shower like (' + r'$\nu_{e}-CC$)'},
               'elec-NC': {'title': 'Track like (' + r'$\nu_{e}-NC$)'},
               'tau-CC': {'title': 'Tau like (' + r'$\nu_{\tau}-CC$)'}}

    fig, ax = plt.subplots()
    pdf_plots = mpl.backends.backend_pdf.PdfPages(savefolder + '/' + savename + '.pdf')

    for ic in ic_list.keys():
        is_ic = select_ic(mc_info['particle_type'], mc_info['is_cc'], ic)
        if bool(np.any(is_ic, axis=0)) is False:
            continue

        bins_prop_1, bins_prop_2 = properties[prop_1_name]['bins'], properties[prop_2_name]['bins']

        hist_2d_prop_1_prop_2 = np.histogram2d(prop_1[is_ic], prop_2[is_ic], [bins_prop_1, bins_prop_2])
        bin_edges_prop_1, bin_edges_prop_2 = hist_2d_prop_1_prop_2[1], hist_2d_prop_1_prop_2[2]

        # Format in classical numpy convention: x along first dim (vertical), y along second dim (horizontal)
        # transpose to get typical cartesian convention: y along first dim (vertical), x along second dim (horizontal)
        pcm_prop_1_prop_2 = ax.pcolormesh(bin_edges_prop_1, bin_edges_prop_2, hist_2d_prop_1_prop_2[0].T,
                                          norm=mpl.colors.LogNorm(vmin=1, vmax=hist_2d_prop_1_prop_2[0].T.max()))

        title_name_prop_1, title_name_prop_2 = properties[prop_1_name]['title_name'], properties[prop_2_name]['title_name']

        plot_line_through_the_origin(prop_1_name, prop_2_name)
        title = plt.title(title_prefix + ic_list[ic]['title'] + ', ' + title_name_prop_1 + ' vs. ' + title_name_prop_2)
        title.set_position([.5, 1.04])
        cbar = fig.colorbar(pcm_prop_1_prop_2, ax=ax)
        cbar.ax.set_ylabel('Number of events')
        x_label, y_label = properties[prop_1_name]['ax_label'], properties[prop_2_name]['ax_label']
        ax.set_xlabel(x_label), ax.set_ylabel(y_label)
        plt.tight_layout()

        pdf_plots.savefig(fig)
        cbar.remove()  # TODO still necessary?
        ax.cla()

    pdf_plots.close()
    plt.close()


def get_make_2d_energy_resolution_plot_properties_dict():
    """
    Returns a dict with a list of all properties that should work with the make_2d_energy_resolution_plot_properties
    function.

    Returns
    -------
    properties : dict
        Dict that contains the configurations of all properties that work with the
        make_2d_energy_resolution_plot_properties function.

    """
    properties = {
                  'energy_reco': {'dset_key': 'pred', 'col_name': 'pred_energy', 'bins': np.arange(1, 101, 1),
                                  'title_name': 'reco energy', 'ax_label': 'Reconstructed energy (GeV)'},
                  'energy_true': {'dset_key': 'mc_info', 'col_name': 'energy', 'bins': np.arange(1, 101, 1),
                                  'title_name': 'true energy', 'ax_label': 'True energy (GeV)'},
                  'dir_x_reco': {'dset_key': 'pred', 'col_name': 'pred_dir_x', 'bins': np.linspace(-1, 1, 100),
                                 'title_name': 'reco dir-x', 'ax_label': 'Reconstructed dir-x'},
                  'dir_y_reco': {'dset_key': 'pred', 'col_name': 'pred_dir_y', 'bins': np.linspace(-1, 1, 100),
                                 'title_name': 'reco dir-y', 'ax_label': 'Reconstructed dir-y'},
                  'dir_z_reco': {'dset_key': 'pred', 'col_name': 'pred_dir_z', 'bins': np.linspace(-1, 1, 100),
                                 'title_name': 'reco dir-z', 'ax_label': 'Reconstructed dir-z'},
                  'dir_x_true': {'dset_key': 'mc_info', 'col_name': 'dir_x', 'bins': np.linspace(-1, 1, 100),
                                 'title_name': 'true dir-x', 'ax_label': 'True dir-x'},
                  'dir_y_true': {'dset_key': 'mc_info', 'col_name': 'dir_y', 'bins': np.linspace(-1, 1, 100),
                                 'title_name': 'true dir-y', 'ax_label': 'True dir-y'},
                  'dir_z_true': {'dset_key': 'mc_info', 'col_name': 'dir_z', 'bins': np.linspace(-1, 1, 100),
                                 'title_name': 'true dir-z', 'ax_label': 'True dir-z'},
                  'azimuth_reco': {'dset_key': 'pred', 'bins': np.linspace(-math.pi, math.pi, 100),
                                   'title_name': 'reco azimuth', 'ax_label': 'Reconstructed azimuth (rad)'},
                  'azimuth_true': {'dset_key': 'mc_info', 'bins': np.linspace(-math.pi, math.pi, 100),
                                   'title_name': 'true azimuth', 'ax_label': 'True azimuth (rad)'},
                  'zenith_reco': {'dset_key': 'pred', 'bins': np.linspace(-math.pi/float(2), math.pi/float(2), 100),
                                  'title_name': 'reco zenith', 'ax_label': 'Reconstructed zenith (rad)'},
                  'zenith_true': {'dset_key': 'mc_info', 'bins': np.linspace(-math.pi/float(2), math.pi/float(2), 100),
                                  'title_name': 'true zenith', 'ax_label': 'True zenith (rad)'},
                  'bjorkeny_reco': {'dset_key': 'pred', 'col_name': 'pred_bjorkeny', 'bins': np.linspace(0, 1, 101),
                                    'title_name': 'reco bjorkeny', 'ax_label': 'Reconstructed bjorkeny'},
                  'bjorkeny_true': {'dset_key': 'mc_info', 'col_name': 'bjorkeny', 'bins': np.linspace(0, 1, 101),
                                    'title_name': 'true bjorkeny', 'ax_label': 'True bjorkeny'},
                  'vtx_x_reco': {'dset_key': 'pred', 'col_name': 'pred_vtx_x', 'bins': 50,
                                 'title_name': 'reco vtx-x', 'ax_label': 'Reconstructed vtx-x'},
                  'vtx_y_reco': {'dset_key': 'pred', 'col_name': 'pred_vtx_y', 'bins': 50,
                                 'title_name': 'reco vtx-y', 'ax_label': 'Reconstructed vtx-y'},
                  'vtx_z_reco': {'dset_key': 'pred', 'col_name': 'pred_vtx_z', 'bins': 50,
                                 'title_name': 'reco vtx-z', 'ax_label': 'Reconstructed vtx-z'},
                  'vtx_x_true': {'dset_key': 'mc_info', 'col_name': 'vtx_x', 'bins': 50,
                                 'title_name': 'true vtx-x', 'ax_label': 'True vtx-x'},
                  'vtx_y_true': {'dset_key': 'mc_info', 'col_name': 'vtx_y', 'bins': 50,
                                 'title_name': 'true vtx-y', 'ax_label': 'True vtx-y'},
                  'vtx_z_true': {'dset_key': 'mc_info', 'col_name': 'vtx_z', 'bins': 50,
                                 'title_name': 'true vtx-z', 'ax_label': 'True vtx-z'}
                  }

    return properties


def get_property_info_to_plot(mc_info, prop_dset, properties, prop_name, reco_energy_correction=None):
    """
    Returns a 1d array which contains the data of the specified property, that should be plotted on one axis.

    Parameters
    ----------
    mc_info : h5py.dataset.Dataset/ndarray(ndim=2)
        The mc_info structured array of an OrcaNet nn prediction file.
    prop_dset : h5py.dataset.Dataset/ndarray(ndim=2)
        The dataset, which contains the property as one column. Coincidentally, can also be equal to the
        mc_info dataset, e.g. if prop_name = 'energy_true'
    properties : dict
        Dict that contains the configurations of all properties that work with the
        make_2d_energy_resolution_plot_properties function.
    prop_name : str
        Name of the property.
    reco_energy_correction : None/str
        If on of the props is the reco energy, specifies if the energy should be corrected or not.
        If not None, the string specifies the metric that should be used for the correction.
        Available: 'median' and 'mean'.

    Returns
    -------
    prop : ndarray(ndim=1)
        1d array which contains the data of the specified property that will be plotted on one axis later on.

    """
    if 'azimuth' in prop_name or 'zenith' in prop_name:
        # atan2(y,x)
        dset_key = properties[prop_name]['dset_key']
        col_name_prefix = '' if dset_key != 'pred' else 'pred_'

        azimuth = convert_vectorial_to_spherical_dir(prop_dset, 'azimuth', col_name_prefix)
        prop = azimuth

    elif 'zenith' in prop_name:
        # atan2(z, sqrt(x**2 + y**2))
        dset_key = properties[prop_name]['dset_key']
        col_name_prefix = '' if dset_key != 'pred' else 'pred_'

        zenith = convert_vectorial_to_spherical_dir(prop_dset, 'zenith', col_name_prefix)
        prop = zenith

    elif 'bjorkeny_true' in prop_name:
        # correct true by to 1 for e-NC events for the plotting
        abs_particle_type, is_cc = np.abs(mc_info['particle_type']), mc_info['is_cc']
        is_e_nc = np.logical_and(abs_particle_type == 12, is_cc == 0)

        col_name = properties[prop_name]['col_name']  # should be bjorkeny
        # here, prop array is actually the column bjorkeny of the mc_info array
        prop = np.copy(prop_dset[col_name])  # TODO test if this really works or if this is actually necessary
        prop[is_e_nc] = 1

    elif prop_name == 'energy_reco':
        col_name = properties[prop_name]['col_name']

        if reco_energy_correction is not None:
            energy_pred_array = prop_dset[col_name]
            energy_pred = correct_reco_energy(mc_info, energy_pred_array, metric=reco_energy_correction)
            prop = energy_pred
        else:
            prop = prop_dset[col_name]

    else:
        # energy_true, dir_x, dir_y, dir_z, vtx_x, ...
        col_name = properties[prop_name]['col_name']
        prop = prop_dset[col_name]

    return prop

# --------------------------- Code for 2d plots --------------------------- #

# --------------------------- Code for 1d plots --------------------------- #


def make_1d_property_errors_metric_over_energy(pred_file, property_name, mode, savefolder, savename,
                                               reco_energy_correction=None, energy_bins=np.linspace(1, 100, 32),
                                               cuts=None, compare_2nd_reco=None):
    """
    Makes binned 1d plots that show
    1) X-axis: The true energy binned into bins
    2) Y-axis: Some performance metric, dependent on the mode[0] argument.

    If mode[0] == 'rel_std_div':
        Plots the relative standard deviation (relative to the e_mc energy bin) of the property
        specified by the property_name
    If mode[0] != 'rel_std_div'
        Plots the reco error residuals of prop_reco - prop_true, binned in energy with a certain metric
        (e.g. mae, median) for the specified interaction channel. The 2nd tuple element mode[1] is the metric
        and must be one of "mae", "median", "mean_relative" or "median_relative".

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the regression predictions of a nn model.
    property_name : str
        Name of the property that should be plotted on the Y-axis, e.g. 'energy'.
        All available properties can be found in the property dict below.
    mode : tuple(str, str)
        The description of this parameter can be found above in the docs of the main function.
        If mode[0] != 'rel_std_div', mode[1] must be one of "mae", "median", "mean_relative" or "median_relative".
        In this case, what is given in mode[0] doesn't matter at all.
        Likewise, if mode[0] == 'rel_std_div', what is specified in mode[1] doesn't matter.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    savename : str
        Savename with which the plots file should be saved.
    reco_energy_correction : None/str
        If the prop is the reco energy, specifies if the energy should be corrected or not.
        If not None, the string specifies the metric that should be used for the correction.
        Available: 'median' and 'mean'.
    energy_bins : ndarray(ndim=1)
        Energy bins that should be used for the binning of the X-axis.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.
    compare_2nd_reco : None/h5py.File
        Either None or h5py.File instance, if the info of a 2nd reco should be plotted into the same plots.
        The file should have the same format as the pred_file.

    """
    fig, ax = plt.subplots()
    pdf_plots = mpl.backends.backend_pdf.PdfPages(savefolder + '/' + savename + '.pdf')
    title_prefix = 'OrcaNet: '

    properties = {'dirs_vector': {'sub_props': ['dir_x', 'dir_y', 'dir_z'], 'ylabel': ' error (dir)'},
                  'dirs_spherical': {'sub_props': ['azimuth', 'zenith'], 'ylabel': ' error (dir)'},
                  'energy': {'sub_props': ['energy'], 'ylabel': ' error (corrected energy)', 'correct': 'median'},
                  'vertex_vector': {'sub_props': ['vtx_x', 'vtx_y', 'vtx_z'], 'ylabel': ' error (vertex)'},
                  'bjorkeny': {'sub_props': ['bjorkeny'], 'ylabel': ' error (bjorkeny)'}}

    ic_list = {'muon-CC': {'title': 'Track like (' + r'$\nu_{\mu}-CC$)'},
               'elec-CC': {'title': 'Shower like (' + r'$\nu_{e}-CC$)'},
               'elec-NC': {'title': 'Track like (' + r'$\nu_{e}-NC$)'},
               'tau-CC': {'title': 'Tau like (' + r'$\nu_{\tau}-CC$)'}}

    mc_info, pred = get_mc_info_and_pred_datasets(pred_file, 'mc_info', 'pred', cuts)
    if compare_2nd_reco is not None:
        pred_file_2 = compare_2nd_reco
        mc_info_2, pred_2 = get_mc_info_and_pred_datasets(pred_file_2, 'mc_info', 'pred', cuts)

    sub_props_list = properties[property_name]['sub_props']
    for ic in ic_list.keys():
        is_ic = select_ic(mc_info['particle_type'], mc_info['is_cc'], ic)
        if bool(np.any(is_ic, axis=0)) is False:
            continue

        for sub_prop in sub_props_list:
            plot_data = calc_plot_data_of_energy_dependent_label(mc_info, pred, sub_prop, mode,
                                                                 energy_bins=energy_bins,
                                                                 reco_energy_correction=reco_energy_correction,)

            bins, perf = plot_data[0], plot_data[1]
            ax.step(bins, perf, linestyle="-", where='post', label='DL ' + sub_prop)

            if compare_2nd_reco is not None:
                # no energy correction for the 2nd file, typically some standard reco
                plot_data_2 = calc_plot_data_of_energy_dependent_label(mc_info_2, pred_2, sub_prop, mode,
                                                                       reco_energy_correction=None,
                                                                       energy_bins=energy_bins)

                ax.step(bins, plot_data_2[1], linestyle="-", where='post', label='Std ' + sub_prop)

        x_ticks_major = np.arange(0, 101, 10)
        ax.set_xticks(x_ticks_major)
        ax.minorticks_on()
        title = plt.title(title_prefix + ic_list[ic]['title'])
        title.set_position([.5, 1.04])
        y_label = r'$\sigma / E_{true}$ for ' + property_name + ' reco' if mode[0] == 'rel_std_div' else mode[1] + properties[property_name]['ylabel']
        ax.set_xlabel('True energy (GeV)'), ax.set_ylabel(y_label)
        ax.grid(True)
        ax.legend(loc='upper right')

        pdf_plots.savefig(fig)
        ax.cla()

    plt.close()
    pdf_plots.close()


def get_mc_info_and_pred_datasets(pred_file, mc_info_key, pred_key, cuts):
    """
    Gets the mc_info and pred_datasets from a h5py pred_file instance and applies some cuts if cuts is not None.

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the regression predictions of a nn model.
    mc_info_key : str
        Key of the mc_info dataset in the pred_file.
    pred_key : str
        Key of the pred dataset in the pred_file.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.

    Returns
    -------
    mc_info : h5py.dataset.Dataset/ndarray(ndim=2)
        The (cut) mc_info structured array of an OrcaNet nn prediction file.
    pred : h5py.dataset.Dataset/ndarray(ndim=2)
        The (cut) pred dataset of a nn OrcaNet classifier, which contains all predicted labels as single columns.

    """
    mc_info = pred_file[mc_info_key]
    pred = pred_file[pred_key]
    if cuts is not None:
        assert isinstance(cuts, str)
        evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
        mc_info = mc_info[evt_sel_mask]
        pred = pred[evt_sel_mask]

    return mc_info, pred


def calc_plot_data_of_energy_dependent_label(mc_info, pred, prop_name, mode,
                                             energy_bins=np.linspace(1, 100, 20), reco_energy_correction=None):
    """
    Returns two different kind of performance metrics vs energy, dependent on the mode parameter:

    1) if mode[0] == 'rel_std_div':
       Gets the relative standard deviation for each bin, which are specified by the energy_bins parameter.
    2) if mode[0] != 'rel_std_div'
        Returns the reco error residuals of prop_reco - prop_true, binned in energy with a certain metric
        (e.g. mae, median). The 2nd tuple element mode[1] is the metric and must be one of
        "mae", "median", "mean_relative" or "median_relative".

    Parameters
    ----------
    mc_info : h5py.dataset.Dataset/ndarray(ndim=2)
        The mc_info structured array of an OrcaNet nn prediction file.
    pred : h5py.dataset.Dataset/ndarray(ndim=2)
        The pred dataset of a nn OrcaNet classifier, which contains all predicted labels as single columns.
    prop_name : str
        Name of the property, of which the reco error residuals should be calculated, e.g. "energy".
    mode : tuple(str, str)
        The description of this parameter can be found above in the docs of the main function.
        If mode[0] != 'rel_std_div', mode[1] must be one of "mae", "median", "mean_relative" or "median_relative".
        In this case, what is given in mode[0] doesn't matter at all.
        Likewise, if mode[0] == 'rel_std_div', what is specified in mode[1] doesn't matter.
    energy_bins : ndarray(ndim=1)
        Energy bins that should be used for the binning.
    reco_energy_correction : None/str
        If the prop is the reco energy, specifies if the energy should be corrected or not.
        If not None, the string specifies the metric that should be used for the correction.
        Available: 'median' and 'mean'.

    Returns
    -------
    energy_to_property_performance_plot_data_ic : tuple(ndarray, ndarray, optional(ndarray))
        If mode[0] == 'rel_std_div':
            Tuple containing 2 arrays:
            1) energy_bins array from the input.
            2) the relative (relative to e-mc-bin) standard deviation of the property for each bin.
        If mode[0] != 'rel_std_div':
            Tuple containing 2, or in case of 'median_relative' 3, arrays:
            1) energy_bins array from the input.
            2) hist_energy_residuals_metric array, which contains the metric value for each energy bin.
            3) hist_energy_variance array, same as 2 but this time the variance.

    """
    if 'azimuth' in prop_name:
        # atan2(y,x)
        azimuth_pred = convert_vectorial_to_spherical_dir(pred, 'azimuth', 'pred_')
        azimuth_true = convert_vectorial_to_spherical_dir(mc_info, 'azimuth', 'pred_')

        prop_pred, prop_true = azimuth_pred, azimuth_true

    elif 'zenith' in prop_name:
        # atan2(z, sqrt(x**2 + y**2))
        zenith_pred = convert_vectorial_to_spherical_dir(pred, 'zenith', 'pred_')
        zenith_true = convert_vectorial_to_spherical_dir(mc_info, 'zenith', 'pred_')

        prop_pred, prop_true = zenith_pred, zenith_true

    elif 'energy' in prop_name:
        if reco_energy_correction is not None:
            energy_pred_array = pred['pred_energy']
            energy_pred = correct_reco_energy(mc_info, energy_pred_array, metric=reco_energy_correction)
            energy_true = mc_info[prop_name]
            prop_pred, prop_true = energy_pred, energy_true
        else:
            prop_pred, prop_true = pred['pred_' + prop_name], mc_info[prop_name]

    elif 'bjorkeny' in prop_name:
        # correct by to 1 for e-NC events
        abs_particle_type, is_cc = np.abs(mc_info['particle_type']), mc_info['is_cc']
        is_e_nc = np.logical_and(abs_particle_type == 12, is_cc == 0)

        bjorkeny_pred = np.copy(pred['pred_bjorkeny'])  # TODO test if this really works
        bjorkeny_pred[is_e_nc] = 1

        bjorkeny_true = mc_info['bjorkeny']

        prop_pred, prop_true = bjorkeny_pred, bjorkeny_true

    else:
        prop_pred, prop_true = pred['pred_' + prop_name], mc_info[prop_name]

    energy_true = mc_info['energy']
    if mode[0] == 'rel_std_div':
        energy_to_property_performance_plot_data_ic = get_rel_std_div_plot_data(prop_pred, energy_true, energy_bins)
    else:
        metric = mode[1]
        err = np.abs(prop_pred - prop_true)
        energy_to_property_performance_plot_data_ic = bin_error_in_energy_bins(err, energy_true, energy_bins, metric=metric)

    return energy_to_property_performance_plot_data_ic


def get_rel_std_div_plot_data(prop_pred, energy_true, energy_bins):
    """
    Function that calculates the relative standard deviation of prop_pred in a certain energy range and returns
    it together with the energy_bins array from the input.

    Parameters
    ----------
    prop_pred : ndarray(ndim=1)
        Prediction values of a property like e.g. energy.
    energy_true : ndarray(ndim=1)
        True MC energy values.
    energy_bins : ndarray(ndim=1)
        Array which specifies the energy bins that should be used for the x-binning.

    Returns
    -------
    energy_binned_rel_std_plot_data : tuple(ndarray, ndarray)
        Tuple containing 2 arrays:
        1) energy_bins array from the input.
        2) the relative (relative to e-mc-bin) standard deviation of the property for each bin.

    """
    std_rel = []  # y-axis data
    for i in range(energy_bins.shape[0] - 1):
        e_range_low, e_range_high = energy_bins[i], energy_bins[i+1]
        e_range_mean = (e_range_low + e_range_high) / float(2)

        # cut out prop_pred values within a certain energy bin
        prop_pred_cut_boolean = np.logical_and(e_range_low < energy_true, energy_true <= e_range_high)
        prop_pred_cut = prop_pred[prop_pred_cut_boolean]

        std_temp = np.std(prop_pred_cut)
        std_rel.append(std_temp / float(e_range_mean))

    # fix for mpl
    std_rel.append(std_rel[-1])
    energy_binned_rel_std_plot_data = [energy_bins, std_rel]

    return energy_binned_rel_std_plot_data


def bin_error_in_energy_bins(err, mc_energy, energy_bins, metric='median_relative'):
    """
    Bins the err residuals into energy bins, and calculates the value of each bin based on a specified metric.

    Parameters
    ----------
    err : ndarray(ndim=1)
        Array with the absolute residual reco - true.
    mc_energy : ndarray(ndim=1)
        Array with the true mc energy.
    energy_bins : ndarray(ndim=1)
        Array which specifies the energy bins that should be used for the x-binning.
    metric : str
        Metric that should be used for calculating the values for each bin.
        One of "mae", "median", "mean_relative", "median_relative".

    Returns
    -------
    energy_binned_err_plot_data : tuple(ndarray, ndarray, optional(ndarray))
        Tuple containing 2, or in case of 'median_relative' or 'mean_relative' 3, arrays:
        1) energy_bins array from the input.
        2) hist_energy_residuals_metric array, which contains the metric value for each energy bin.
        3) hist_energy_variance array, same as 2 but this time the variance.

    """
    # bin the err, depending on their mc_energy, into energy_bins
    hist_energy_residuals_metric = np.zeros((len(energy_bins) - 1))
    hist_energy_variance = np.zeros((len(energy_bins) - 1))
    # In which bin does each event belong, according to its mc energy:
    bin_indices = np.digitize(mc_energy, bins=energy_bins)

    # For every mc energy bin, calculate the merge metric (e.g. mae or median relative) of all events that have a corresponding mc energy
    for bin_no in range(1, len(energy_bins)):
        current_err = err[bin_indices == bin_no]
        current_mc_energy = mc_energy[bin_indices == bin_no]

        if metric == 'mae':
            # calculate mean absolute error
            hist_energy_residuals_metric[bin_no - 1] = np.mean(current_err)
        elif metric == 'median_relative':
            # calculate the median of the relative error: |label_reco-label_true|/E_true and also its variance
            # probably makes only sense if the label from the err array is the energy
            relative_error = current_err / current_mc_energy
            hist_energy_residuals_metric[bin_no - 1] = np.median(relative_error)
            hist_energy_variance[bin_no - 1] = np.var(relative_error)
        elif metric == 'mean_relative':
            # calculate the median of the relative error: |label_reco-label_true|/E_true and also its variance
            # probably makes only sense if the label from the err array is the energy
            relative_error = current_err / current_mc_energy
            hist_energy_residuals_metric[bin_no - 1] = np.mean(relative_error)
            hist_energy_variance[bin_no - 1] = np.var(relative_error)
        elif metric == 'median':
            hist_energy_residuals_metric[bin_no - 1] = np.median(current_err)
        else:
            raise ValueError('Operation modes other than "mae" and "median_relative" are not supported.')

    # For proper plotting with plt.step and where="post"
    hist_energy_residuals_metric = np.append(hist_energy_residuals_metric, hist_energy_residuals_metric[-1])
    hist_energy_variance = np.append(hist_energy_variance, hist_energy_variance[-1]) if metric in ['median_relative', 'mean_relative'] else None
    energy_binned_err_plot_data = [energy_bins, hist_energy_residuals_metric, hist_energy_variance]

    return energy_binned_err_plot_data

# --------------------------- Code for 1d plots --------------------------- #


# --------------------------- Code for error plots --------------------------- #


def make_1d_reco_err_div_by_std_dev_plot(pred_file, savefolder, savename, cuts=None):
    """
    Makes a 1d hist plot which shows the reco err (y_true - y_reco) divided
    by the predicted standard deviation error of a nn.

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the regression predictions of a nn model.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    savename : str
        Savename with which the plots file should be saved.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.

    """
    # do this for all labels: 1d (y_true - y_reco) / std_reco
    fig, ax = plt.subplots()
    pdf_plots = mpl.backends.backend_pdf.PdfPages(savefolder + '/' + savename + '.pdf')

    prop_names = {'energy': {'col_name_pred': 'pred_energy', 'col_name_pred_err': 'pred_err_energy', 'col_name_true': 'true_energy'},
                  'bjorkeny': {'col_name_pred': 'pred_bjorkeny', 'col_name_pred_err': 'pred_err_bjorkeny', 'col_name_true': 'true_bjorkeny'},
                  'dir_x': {'col_name_pred': 'pred_dir_x', 'col_name_pred_err': 'pred_err_dir_x', 'col_name_true': 'true_dir_x'},
                  'dir_y': {'col_name_pred': 'pred_dir_y', 'col_name_pred_err': 'pred_err_dir_y', 'col_name_true': 'true_dir_y'},
                  'dir_z': {'col_name_pred': 'pred_dir_z', 'col_name_pred_err': 'pred_err_dir_z', 'col_name_true': 'true_dir_z'},
                  'vtx_x': {'col_name_pred': 'pred_vtx_x', 'col_name_pred_err': 'pred_err_vtx_x', 'col_name_true': 'true_vxt_x'},
                  'vtx_y': {'col_name_pred': 'pred_vtx_y', 'col_name_pred_err': 'pred_err_vtx_y', 'col_name_true': 'true_vxt_y'},
                  'vtx_z': {'col_name_pred': 'pred_vtx_z', 'col_name_pred_err': 'pred_err_vtx_z', 'col_name_true': 'true_vxt_z'},
                  'vtx_t': {'col_name_pred': 'pred_vtx_t', 'col_name_pred_err': 'pred_err_vtx_t', 'col_name_true': 'true_vxt_t'}}

    mc_info = pred_file['mc_info']
    for prop in prop_names:
        prop_pred = pred_file['pred'][prop_names[prop]['col_name_pred']]
        prop_pred_err = pred_file['pred'][prop_names[prop]['col_name_pred_err']]
        prop_true = pred_file['true'][prop_names[prop]['col_name_true']]

        if cuts is not None:
            assert isinstance(cuts, str)
            evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
            prop_pred = prop_pred[evt_sel_mask]
            prop_pred_err = prop_pred_err[evt_sel_mask]
            prop_true = prop_true[evt_sel_mask]

        plot_1d_reco_err_div_by_std_for_label(prop_pred, prop_pred_err, prop_true, fig, ax, pdf_plots, prop)

    plt.close()
    pdf_plots.close()


def plot_1d_reco_err_div_by_std_for_label(prop_pred, prop_pred_err, prop_true, fig, ax, pdf_plots, label):
    """
    Plots the reco err (y_true - y_reco) divided by the predicted standard deviation error
    to an ax of a fig and saves them to the pdf_plots instance.

    Parameters
    ----------
    prop_pred : ndarray(ndim=1)
        Array with the nn predictions of a certain property (e.g. pred_energy).
    prop_pred_err : ndarray(ndim=1)
        Array with the nn error predictions of a certain property.
    prop_true : ndarray(ndim=1)
        Array with the true values of a certain property (e.g. pred_energy).
    fig : mpl figure
        Matplotlib instance of a figure.
    ax : mpl.axes
        Axes object that refers to ax of an existing plt.sublots object.
    pdf_plots : mpl.backends.backend_pdf.PdfPages
        Instance of a matplotlib PdfPages object, to which the plot should be saved.
    label : str
        Label that should be used as a str in the plot with plt.hist.

    """
    # convert mse error to standard deviation with the magic number sqrt(pi/2)
    prop_pred_err_std = prop_pred_err * 1.253

    pred_err_div_by_std = np.divide(prop_true - prop_pred, np.abs(prop_pred_err_std))
    exclude_outliers = np.logical_and(pred_err_div_by_std > -10, pred_err_div_by_std < 10)

    # print(np.std(pred_err_div_by_std[exclude_outliers]))
    plt.hist(pred_err_div_by_std[exclude_outliers], bins=100, label=label)

    title = plt.title('Gaussian Likelihood errors for ' + label)
    title.set_position([.5, 1.04])
    ax.set_xlabel(r'$(y_{\mathrm{true}} - y_{\mathrm{pred}})/ \sigma_{\mathrm{pred}}$'), ax.set_ylabel('Counts [#]')
    ax.grid(True)

    pdf_plots.savefig(fig)
    ax.cla()


def make_1d_reco_err_to_reco_residual_plot(pred_file, savefolder, savename, cuts=None):
    """
    Makes a 1d plot which shows the true standard deviation of the error residuals (abs(true-pred)) versus
    the estimated uncertainty sigma_pred by a nn.

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the regression predictions of a nn model.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    savename : str
        Savename with which the plots file should be saved.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.

    """
    fig, ax = plt.subplots()
    pdf_plots = mpl.backends.backend_pdf.PdfPages(savefolder + '/' + savename + '.pdf')

    ic_list = {'muon-CC': {'title': 'Track like (' + r'$\nu_{\mu}-CC$)'},
               'elec-CC': {'title': 'Shower like (' + r'$\nu_{e}-CC$)'},
               'elec-NC': {'title': 'Track like (' + r'$\nu_{e}-NC$)'},
               'tau-CC': {'title': 'Tau like (' + r'$\nu_{\tau}-CC$)'}}

    properties = {'energy': {'col_name_pred': 'pred_energy', 'col_name_true': 'true_energy', 'col_name_pred_err': 'pred_err_energy', 'unit': '[GeV]'},
                  'bjorkeny': {'col_name_pred': 'pred_bjorkeny', 'col_name_true': 'true_bjorkeny', 'col_name_pred_err': 'pred_err_bjorkeny', 'unit': ''},
                  'dir_x': {'col_name_pred': 'pred_dir_x', 'col_name_true': 'true_dir_x', 'col_name_pred_err': 'pred_err_dir_x', 'unit': '[rad]'},
                  'dir_y': {'col_name_pred': 'pred_dir_y', 'col_name_true': 'true_dir_y', 'col_name_pred_err': 'pred_err_dir_y', 'unit': '[rad]'},
                  'dir_z': {'col_name_pred': 'pred_dir_z', 'col_name_true': 'true_dir_z', 'col_name_pred_err': 'pred_err_dir_z', 'unit': '[rad]'},
                  'azimuth': {'col_name_pred': None, 'col_name_true': None, 'col_name_pred_err': None, 'unit': '[rad]'},
                  'zenith': {'col_name_pred': None, 'col_name_true': None, 'col_name_pred_err': None, 'unit': '[rad]'},
                  'vtx_x': {'col_name_pred': 'pred_vtx_x', 'col_name_true': 'true_vtx_x', 'col_name_pred_err': 'pred_err_vtx_x', 'unit': '[rad]'},
                  'vtx_y': {'col_name_pred': 'pred_vtx_y', 'col_name_true': 'true_vtx_y', 'col_name_pred_err': 'pred_err_vtx_y', 'unit': '[rad]'},
                  'vtx_z': {'col_name_pred': 'pred_vtx_z', 'col_name_true': 'true_vtx_z', 'col_name_pred_err': 'pred_err_vtx_z', 'unit': '[rad]'},
                  'vtx_t': {'col_name_pred': 'pred_vtx_t', 'col_name_true': 'true_vtx_t', 'col_name_pred_err': 'pred_err_vtx_t', 'unit': '[rad]'}}

    mc_info, pred, true = pred_file['mc_info'], pred_file['pred'], pred_file['true']
    if cuts is not None:
        assert isinstance(cuts, str)
        evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
        mc_info = mc_info[evt_sel_mask]
        pred = pred[evt_sel_mask]
        true = true[evt_sel_mask]

    for prop_name in properties:
        for ic in ic_list:
            is_ic = select_ic(mc_info['particle_type'], mc_info['is_cc'], ic)
            if bool(np.any(is_ic, axis=0)) is False:
                continue

            mc_info_ic = mc_info[is_ic]
            pred_ic, true_ic = pred[is_ic], true[is_ic]
            prop_true, prop_pred, prop_pred_err = get_prop_true_pred_and_prop_pred_sigma(pred_ic, true_ic, mc_info_ic,
                                                                                         properties, prop_name)

            n_x_bins = 50
            title = plt.title(ic_list[ic]['title'] + ': ' + prop_name)
            unit = properties[prop_name]['unit']

            plot_1d_reco_err_to_reco_residual_for_prop(prop_true, prop_pred, prop_pred_err, n_x_bins, fig,
                                                       ax, title, pdf_plots, prop_name, unit)

    plt.close()
    pdf_plots.close()


def get_prop_true_pred_and_prop_pred_sigma(pred, true, mc_info, properties, prop_name):
    """
    Gets the data that is needed to plot the true standard deviation of the error residuals (abs(true-pred)) versus
    the estimated uncertainty sigma_pred by a nn.

    Parameters
    ----------
    pred : h5py.dataset.Dataset/ndarray(ndim=2)
        The pred dataset of an OrcaNet nn prediction file.
    true : h5py.dataset.Dataset/ndarray(ndim=2)
        The true dataset of an OrcaNet nn prediction file.
    mc_info : h5py.dataset.Dataset/ndarray(ndim=2)
        The mc_info dataset of an OrcaNet nn prediction file.
    properties : dict
        Dict which contains some configurations of all properties.
    prop_name : str
        String which specifies the specific property that should be used.

    Returns
    -------
    prop_true : ndarray(ndim=1)
        Array which contains the true values of the specified property.
    prop_pred : ndarray(ndim=1)
        Array which contains the predicted values of the specified property.
    prop_pred_err : ndarray(ndim=1)
        Array which contains the predicted error values of the specified property, multiplied by the magic number
        sqrt(pi/2) in order to get a standard deviation from the error values if trained with mse loss.
        If the prop is azimuth or zenith, the error of the vectorial dirs is propagated with error propagation
        that neglects the correlation term.

    """

    if prop_name == 'azimuth' or prop_name == 'zenith':

        dx_true, dx_pred = true[properties['dir_x']['col_name_true']], pred[properties['dir_x']['col_name_pred']]
        dy_true, dy_pred = true[properties['dir_y']['col_name_true']], pred[properties['dir_y']['col_name_pred']]
        dz_true, dz_pred = true[properties['dir_z']['col_name_true']], pred[properties['dir_z']['col_name_pred']]

        dx_pred_err = pred[properties['dir_x']['col_name_pred_err']]
        dy_pred_err = pred[properties['dir_y']['col_name_pred_err']]
        dz_pred_err = pred[properties['dir_z']['col_name_pred_err']]

        correction = 1.253  # convert mse error to standard deviation with the magic number sqrt(pi/2)
        if prop_name == 'azimuth':
            azimuth_true = convert_vectorial_to_spherical_dir(true, 'azimuth', col_name_prefix='')
            azimuth_pred = convert_vectorial_to_spherical_dir(pred, 'azimuth', col_name_prefix='pred_')

            # clip std prediction to 0 (maybe not really necessary) and apply correction
            dx_pred_err = np.clip(np.copy(dx_pred_err), 0, None) * correction  # copy maybe not necessary
            dy_pred_err = np.clip(np.copy(dy_pred_err), 0, None) * correction

            # error propagation of atan2 with correlations between y and x neglected (covariance term = 0)
            azimuth_pred_err = np.sqrt((dx_pred / (dx_pred ** 2 + dy_pred ** 2)) ** 2 * dy_pred_err ** 2 +
                                       (-dy_pred / (dx_pred ** 2 + dy_pred ** 2)) ** 2 * dx_pred_err ** 2)

            # clip every predicted standard deviation to pi if larger than pi
            n = azimuth_pred_err > math.pi
            percentage_true = np.sum(n) / float(azimuth_pred_err.shape[0]) * 100
            print('Clipped ' + str(n) + ' predicted azimuth standard deviations to pi ('
                  + str(percentage_true) + '% of all events)')

            azimuth_pred_err = np.clip(azimuth_pred_err, None, math.pi)

            # # If we don't want to neglect the convariance term , since y and x are correlated:
            # # tested: doesn't really make a difference
            # mean_dx, mean_dy = np.mean(dx_pred), np.mean(dy_pred)
            # covariance_xy = 1/float(dx_pred.shape[0]) * np.sum((dx_pred - mean_dx) * (dy_pred - mean_dy))
            # test = np.abs(2 * (dx_pred / (dx_pred ** 2 + dy_pred ** 2)) * (-dy_pred / (dx_pred ** 2 + dy_pred ** 2)) * covariance_xy)
            # label_std_pred = np.sqrt(np.abs(( dx_pred / (dx_pred ** 2 + dy_pred ** 2)) ** 2 * dy_std_pred ** 2 +
            #                          (-dy_pred / (dx_pred ** 2 + dy_pred ** 2)) ** 2 * dx_std_pred ** 2 +
            #                          2 * (dx_pred / (dx_pred ** 2 + dy_pred ** 2)) * (-dy_pred / (dx_pred ** 2 + dy_pred ** 2)) * covariance_xy))
            # label_std_pred = np.clip(label_std_pred, None, math.pi/float(2))

            prop_true, prop_pred, prop_pred_err = azimuth_true, azimuth_pred, azimuth_pred_err

        else:  # zenith
            zenith_true = convert_vectorial_to_spherical_dir(true, 'zenith', col_name_prefix='')
            zenith_pred = convert_vectorial_to_spherical_dir(pred, 'zenith', col_name_prefix='pred_')

            # clip std prediction to 0 (maybe not really necessary) and apply correction
            dz_pred_err = np.clip(np.copy(dz_pred_err), 0, None) * correction

            # error propagation with correlations neglected (covariance term = 0)
            zenith_pred_err = np.sqrt((-1 / np.sqrt((1 - dz_pred ** 2))) ** 2 * dz_pred_err ** 2)  # zen = arccos(z/norm(r))

            # # If we don't want to neglect the convariance term:
            # dx_std_pred, dy_std_pred = np.clip(dx_std_pred, 0, None) * correction, np.clip(dy_std_pred, 0, None) * correction
            # dx_pred, dy_pred = np.clip(dx_pred, -0.999, 0.999), np.clip(dy_pred, -0.999, 0.999)
            # label_std_pred = np.sqrt( (np.sqrt(dx_pred ** 2 + dy_pred ** 2) / (dx_pred ** 2 + dy_pred ** 2 + dz_pred ** 2))**2 * dz_std_pred ** 2 +
            #                           ((- dx_pred * dz_pred) / (np.sqrt(dx_pred ** 2 + dy_pred ** 2) * (dx_pred ** 2 + dy_pred ** 2 + dz_pred ** 2))) ** 2 * dx_std_pred ** 2 +
            #                           ((- dy_pred * dz_pred) / (np.sqrt(dx_pred ** 2 + dy_pred ** 2) * (dx_pred ** 2 + dy_pred ** 2 + dz_pred ** 2))) ** 2 * dy_std_pred ** 2)

            prop_true, prop_pred, prop_pred_err = zenith_true, zenith_pred, zenith_pred_err

    else:
        if prop_name == 'bjorkeny':
            # correct true by to 1 for e-NC events for the plotting
            abs_particle_type, is_cc = np.abs(mc_info['particle_type']), mc_info['is_cc']
            is_e_nc = np.logical_and(abs_particle_type == 12, is_cc == 0)

            prop_true = np.copy(true[properties[prop_name]['col_name_true']])  # TODO test if this really works or if this is actually necessary
            prop_true[is_e_nc] = 1
        else:
            prop_true = true[properties[prop_name]['col_name_true']]

        prop_pred = true[properties[prop_name]['col_name_pred']]
        prop_pred_err = true[properties[prop_name]['col_name_pred_err']]

        prop_pred_err = prop_pred_err * 1.253  # sqrt(pi/2) magic number correction with mse training
        prop_pred_err = np.abs(prop_pred_err)  # TODO necessary? -> rather set to zero, regarding the loss function?

    return prop_true, prop_pred, prop_pred_err


def plot_1d_reco_err_to_reco_residual_for_prop(prop_true, prop_pred, prop_pred_err, n_x_bins, fig, ax, title,
                                               pdf_plots, prop_name, unit):
    """
    Plots the true standard deviation of the error residuals (abs(true-pred)) versus
    the estimated uncertainty sigma_pred by a nn to a mpl ax and saves it to a mpl PdfPages instance.

    Parameters
    ----------
    prop_true : ndarray(ndim=1)
        Array which contains the true values of the specified property.
    prop_pred : ndarray(ndim=1)
        Array which contains the predicted values of the specified property.
    prop_pred_err : ndarray(ndim=1)
        Array which contains the predicted error values of the specified property, multiplied by the magic number.
    n_x_bins : int
        Number of x-bins.
    fig : mpl figure
        Matplotlib instance of a figure.
    ax : mpl.axes
        Axes object that refers to ax of an existing plt.sublots object.
    title : str
        Title that should be used for the plot.
    pdf_plots : mpl.backends.backend_pdf.PdfPages
        Instance of a matplotlib PdfPages object, to which the plot should be saved.
    prop_name : str
        Name of the property for that this plot should be made.
    unit : str
        String, which specifies the unit of the property and gets appended to the x and y axis labels.

    """
    prop_pred_std_range = (np.amin(prop_pred_err), np.amax(prop_pred_err))
    x_bins_std_pred = np.linspace(prop_pred_std_range[0], prop_pred_std_range[1], n_x_bins + 1)

    x, y = [], []
    for i in range(x_bins_std_pred.shape[0] - 1):  # same as n_x_bins
        prop_std_pred_low, prop_std_pred_high = x_bins_std_pred[i], x_bins_std_pred[i+1]
        prop_std_pred_mean = (prop_std_pred_low + prop_std_pred_high) / float(2)

        prop_std_pred_boolean_mask = np.logical_and(prop_std_pred_low < prop_pred_err, prop_pred_err <= prop_std_pred_high)
        if np.count_nonzero(prop_std_pred_boolean_mask) < 100:
            continue

        if prop_name == 'azimuth':
            # if abs(residual_azimuth) <= pi, take az_true - az_pred ; if residual_azimuth > pi take
            all_residuals = prop_true[prop_std_pred_boolean_mask] - prop_pred[prop_std_pred_boolean_mask]
            larger_pi = np.abs(all_residuals) > math.pi
            sign_larger_pi = np.sign(all_residuals[larger_pi])  # loophole for residual == 0, but doesn't matter below
            all_residuals[larger_pi] = sign_larger_pi * 2 * math.pi - (all_residuals[larger_pi])  # need same sign for 2pi compared to all_residuals value
            # print np.amin(all_residuals), np.amax(all_residuals)

            residuals_std = np.std(all_residuals)

        else:
            residuals_std = np.std(prop_true[prop_std_pred_boolean_mask] - prop_pred[prop_std_pred_boolean_mask])

        x.append(prop_std_pred_mean)
        y.append(residuals_std)

    plt.scatter(x, y, s=20, lw=0.75, c='blue', marker='+')

    title = plt.title(title)
    title.set_position([.5, 1.04])
    ax.set_xlabel(r'Estimated uncertainty $\sigma_{pred}$ ' + unit), ax.set_ylabel('Standard deviation of residuals ' + unit)
    ax.grid(True)

    pdf_plots.savefig(fig)
    ax.cla()


def make_2d_true_reco_plot_different_sigmas(pred_file, savefolder, savename, cuts=None):
    """
    Makes 2d true to reco plots of all properties in the properties dict, and selects only
    the best percentage of events based on the predicted errors by a nn.

    Parameters
    ----------
    pred_file : h5py.File
        H5py file instance, which stores the regression predictions of a nn model.
    savefolder : str
        Path of the directory, where the plots should be saved to.
    savename : str
        Savename with which the plots file should be saved.
    cuts : None/str
        Specifies, if cuts should be used for the plot. Either None or a str, that is available in the
        load_event_selection_file() function.

    """
    fig, ax = plt.subplots()
    pdf_plots = mpl.backends.backend_pdf.PdfPages(savefolder + '/' + savename + '.pdf')

    ic_list = {'muon-CC': {'title': 'Track like (' + r'$\nu_{\mu}-CC$)'},
               'elec-CC': {'title': 'Shower like (' + r'$\nu_{e}-CC$)'},
               'elec-NC': {'title': 'Track like (' + r'$\nu_{e}-NC$)'},
               'tau-CC': {'title': 'Tau like (' + r'$\nu_{\tau}-CC$)'}}

    properties = {'energy': {'bins': np.linspace(1, 100, 100), 'axis_prop_info': ('energy', 'GeV')},
                  'azimuth': {'bins': np.linspace(-math.pi, math.pi, 100), 'axis_prop_info': ('azimuth', 'rad')},
                  'zenith': {'bins': np.linspace(-math.pi/float(2), math.pi/float(2), 100), 'axis_prop_info': ('azimuth', 'rad')}}

    mc_info, pred, true = pred_file['mc_info'], pred_file['pred'], pred_file['true']
    if cuts is not None:
        assert isinstance(cuts, str)
        evt_sel_mask = get_event_selection_mask(mc_info, cut_name=cuts)
        mc_info = mc_info[evt_sel_mask]
        pred = pred[evt_sel_mask]
        true = true[evt_sel_mask]

    for prop_name in properties:
        for ic in ic_list:
            is_ic = select_ic(mc_info['particle_type'], mc_info['is_cc'], ic)
            if bool(np.any(is_ic, axis=0)) is False:
                continue

            mc_info_ic = mc_info[is_ic]
            pred_ic, true_ic = pred[is_ic], true[is_ic]
            prop_true, prop_pred, prop_pred_err = get_prop_true_pred_and_prop_pred_sigma(pred_ic, true_ic, mc_info_ic,
                                                                                         properties, prop_name)

            bins = properties[prop_name]['bins']
            title = ic_list[ic]['title']
            percentage_of_evts = [1.0, 0.8, 0.5, 0.2]

            plot_2d_dir_correlation_different_sigmas(prop_true, prop_pred, prop_pred_err, prop_name, properties,
                                                     percentage_of_evts, title, bins, fig, ax, pdf_plots)

    plt.close()
    pdf_plots.close()


def plot_2d_dir_correlation_different_sigmas(prop_true, prop_pred, prop_pred_err, prop_name, properties,
                                             percentage_of_evts, title, bins, fig, ax, pdf_plots):
    """
    Plots a 2d true to reco plot for a certain property, and selects only
    the best percentage of events based on the predicted errors by a nn.

    Parameters
    ----------
    prop_true : ndarray(ndim=1)
        Array which contains the true values of the specified property.
    prop_pred : ndarray(ndim=1)
        Array which contains the predicted values of the specified property.
    prop_pred_err : ndarray(ndim=1)
        Array which contains the predicted error values of the specified property, multiplied by the magic number.
    prop_name : str
        Name of the property for that this plot should be made.
    properties : dict
        Dict which contains some configurations of all properties.
    percentage_of_evts : list
        List, which contains the percentage of best events. For each value in the list, a separate plot is made.
    title : str
        Title that should be used for the plot.
    bins : ndarray(ndim=1)
        Bins that should be used for both the X and Y axis.
    fig : mpl figure
        Matplotlib instance of a figure.
    ax : mpl.axes
        Axes object that refers to ax of an existing plt.sublots object.
    pdf_plots : mpl.backends.backend_pdf.PdfPages
        Instance of a matplotlib PdfPages object, to which the plot should be saved.

    """
    cbar_max = None
    for i, percentage in enumerate(percentage_of_evts):

        n_total_evt = prop_pred_err.shape[0]
        n_events_to_keep = int(n_total_evt * percentage)  # how many events should be kept

        if prop_name == 'energy':  # TODO needs to be improved, maybe fit line through energyres per escale and divide by this func?
            std_pred_div_by_e_reco = np.divide(prop_pred_err, prop_pred)
            indices_events_to_keep = sorted(std_pred_div_by_e_reco.argsort()[:n_events_to_keep])  # select n minimum values in array

        else:
            indices_events_to_keep = sorted(prop_pred_err.argsort()[:n_events_to_keep])  # select n minimum values in array

        hist_2d_prop = np.histogram2d(prop_true[indices_events_to_keep], prop_pred[indices_events_to_keep], bins)
        bin_edges_prop = hist_2d_prop[1]

        if i == 0:
            cbar_max = hist_2d_prop[0].T.max()

        prop_true_to_reco = ax.pcolormesh(bin_edges_prop, bin_edges_prop, hist_2d_prop[0].T,
                                          norm=mpl.colors.LogNorm(vmin=1, vmax=cbar_max))

        plot_line_through_the_origin(prop_name, prop_name)

        title = plt.title('OrcaNet: ' + title + ', ' + str(int(percentage * 100)) + '% of total events')
        title.set_position([.5, 1.04])
        cbar = fig.colorbar(prop_true_to_reco, ax=ax)
        cbar.ax.set_ylabel('Number of events')

        axis_prop_info = properties[prop_name]['axis_prop_info']
        ax.set_xlabel('True ' + axis_prop_info[0] + ' [' + axis_prop_info[1] + ']')
        ax.set_ylabel('Reconstructed ' + axis_prop_info[0] + ' [' + axis_prop_info[1] + ']')
        plt.xlim(bins[0], bins[-1])
        plt.ylim(bins[0], bins[-1])

        # plt.tight_layout()
        pdf_plots.savefig(fig)

        if i > 0:
            # 1st plot, plot transparency of 100%
            hist_2d_prop_all = np.histogram2d(prop_true, prop_pred, bins)
            bin_edges_prop_all = hist_2d_prop_all[1]

            # only plot the bins that are not anymore in the histogram from above
            hist_2d_prop_larger_zero = np.invert(hist_2d_prop[0] == 0)
            hist_2d_prop_all[0][hist_2d_prop_larger_zero] = 0  # set everything to 0, when these bins are still > 0 in hist_2d_prop

            ax.pcolormesh(bin_edges_prop_all, bin_edges_prop_all, hist_2d_prop_all[0].T, alpha=0.5,
                          norm=mpl.colors.LogNorm(vmin=1, vmax=cbar_max))
            pdf_plots.savefig(fig)
            cbar.remove()
            plt.cla()

            # 2nd plot
            # only divide nonzero bins of hist_2d_prop_all
            hist_2d_prop_all = np.histogram2d(prop_true, prop_pred, bins)
            non_zero = hist_2d_prop_all[0] > 0
            hist_2d_all_div_leftover = np.copy(hist_2d_prop_all[0])
            hist_2d_all_div_leftover[non_zero] = np.divide(hist_2d_prop[0][non_zero], hist_2d_prop_all[0][non_zero])
            # hist_2d_all_div_leftover[np.invert(non_zero)] = -1
            corr_all_div_leftover = ax.pcolormesh(bin_edges_prop_all, bin_edges_prop_all, hist_2d_all_div_leftover.T, vmin=0, vmax=1)
            cbar_2 = fig.colorbar(corr_all_div_leftover, ax=ax)
            cbar_2.ax.set_ylabel('Fraction of leftover events')

            title = plt.title('OrcaNet: ' + title + ', ' + str(int(percentage * 100)) + '% of total events')
            title.set_position([.5, 1.04])
            ax.set_xlabel('True ' + axis_prop_info[0] + ' [' + axis_prop_info[1] + ']')
            ax.set_ylabel('Reconstructed ' + axis_prop_info[0] + ' [' + axis_prop_info[1] + ']')

            pdf_plots.savefig(fig)
            cbar_2.remove()
            plt.cla()

        if i == 0:
            cbar.remove()
        plt.cla()


# --------------------------- Code for error plots --------------------------- #


# --------------------------- Utility code --------------------------- #


def plot_line_through_the_origin(prop_1_name, prop_2_name):
    """
    Plots a line through the origin, if the two propertie names are recognized.

    Parameters
    ----------
    prop_1_name : str
        Name of the first property, e.g. 'energy_reco'.
    prop_2_name : str
        Name of the second property.

    """
    if 'azimuth' in prop_1_name and 'azimuth' in prop_2_name:
        pi = math.pi
        plt.plot([-pi, pi], [-pi, pi], 'k-', lw=1, zorder=10)

    elif 'zenith' in prop_1_name and 'zenith' in prop_2_name:
        pi = math.pi
        plt.plot([-pi/float(2), pi/float(2)], [-pi/float(2), pi/float(2)], 'k-', lw=1, zorder=10)

    elif 'bjorkeny' in prop_1_name and 'bjorkeny' in prop_2_name:
        plt.plot([0, 1], [0, 1], 'k-', lw=1, zorder=10)

    elif 'energy' in prop_1_name and 'energy' in prop_2_name:
        plt.plot([-1, 1], [-1, 1], 'k-', lw=1, zorder=10)

    else:
        warnings.warn('Couldnt decipher how to plot the line through the origin for your two properties'
                      ' that are to plot, ' + prop_1_name + ' & ' + prop_2_name + '.'
                      ' Plotting no line through the origin.')


def convert_vectorial_to_spherical_dir(vect_arr, spherical_coord_name, col_name_prefix=''):
    """
    Returns directions in azimuth / zenith based on an input array with vectorial (dir_x, dir_y, dir_z) coordinates.

    Parameters
    ----------
    vect_arr : ndarray(ndim=2)
        Structured array, which must contain dir_x, dir_y, dir_z
    spherical_coord_name : str
        String which specifies if azimuth or zenith should be returned.
    col_name_prefix : str
        String prefix that should be used for accessing the dir_x/dir_y/dir_z columns.

    Returns
    -------
    azimuth/zenith : ndarray(ndim=1)
        Array with azimuth / zenith spherical directions.

    """
    if spherical_coord_name == 'azimuth':
        # atan2(y,x)

        dir_x = vect_arr[col_name_prefix + 'dir_x']
        dir_y = vect_arr[col_name_prefix + 'dir_y']
        azimuth = np.arctan2(dir_y, dir_x)
        return azimuth

    elif spherical_coord_name == 'zenith':
        # atan2(z, sqrt(x**2 + y**2))

        dir_x = vect_arr[col_name_prefix + 'dir_x']
        dir_y = vect_arr[col_name_prefix + 'dir_y']
        dir_z = vect_arr[col_name_prefix + 'dir_z']

        zenith = np.arctan2(dir_z, np.sqrt(np.power(dir_x, 2) + np.power(dir_y, 2)))
        return zenith

    else:
        raise ValueError('Vectorial to spherical conversions other than azimuth and zenith are not known, you '
                         'specified ' + str(spherical_coord_name))


# --------------------------- Utility code --------------------------- #



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Code for making performance plots based on nn model predictions.
"""

import os
import matplotlib as mpl
mpl.use('Agg')
import h5py
from orcanet_contrib.plotting.bg_classifier import make_prob_hists_bg_classifier
from orcanet_contrib.plotting.ts_classifier import make_e_to_acc_plot_ts, make_ts_prob_hists, plot_ts_separability
from orcanet_contrib.plotting.regression import (make_2d_prop_to_prop_plot,
                                                 make_1d_property_errors_metric_over_energy,
                                                 make_1d_reco_err_div_by_std_dev_plot,
                                                 make_1d_reco_err_to_reco_residual_plot,
                                                 make_2d_true_reco_plot_different_sigmas)


def make_performance_plots(pred_filepath, dataset_modifier, plots_folder):
    """
    Function that makes plots based on the results of a hdf5 file with nn predictions (specified in pred_filepath).

    Parameters
    ----------
    pred_filepath : str
        Path to a h5 OrcaNet prediction file.
    dataset_modifier : str
        String that specifies the dataset modifier that is used for the file located at pred_filepath.
    plots_folder : str
        Path to the general plots folder in the OrcaNet dir structure.

    """
    pred_file = h5py.File(pred_filepath, 'r')
    main_perf_plots_path = plots_folder + '/pred_performance'

    if dataset_modifier == 'bg_classifier':
        cuts = 'bg_classifier'

        make_plots_subfolders(main_perf_plots_path, dataset_modifier)
        make_prob_hists_bg_classifier(pred_file, main_perf_plots_path + '/1d', cuts=cuts)

    elif dataset_modifier == 'ts_classifier':
        cuts = 'neutrino_ts'
        pred_file_2 = None

        make_plots_subfolders(main_perf_plots_path, dataset_modifier)
        make_e_to_acc_plot_ts(pred_file, 'Classified as track', main_perf_plots_path + '/1d', cuts=cuts, prob_threshold_shower=0.5)
        make_ts_prob_hists(pred_file, main_perf_plots_path + '/1d', cuts=cuts)
        plot_ts_separability(pred_file, main_perf_plots_path + '/1d', pred_file_2=pred_file_2, cuts=cuts)

    elif 'regression' in dataset_modifier:
        make_plots_subfolders(main_perf_plots_path, dataset_modifier)
        cuts = 'neutrino_regr'
        reco_energy_correction = 'median'
        pred_file_2nd_reco = None

        if 'energy' in dataset_modifier:
            print('Generating plots for energy performance investigations')
            make_2d_prop_to_prop_plot(pred_file, 'energy_true', 'energy_reco', main_perf_plots_path + '/2d',
                                      'e_true_to_e_reco', cuts=cuts, title_prefix='OrcaNet: ')
            if pred_file_2nd_reco is not None:
                make_2d_prop_to_prop_plot(pred_file_2nd_reco, 'energy_true', 'energy_reco', main_perf_plots_path + '/2d',
                                          'standard_reco_e_true_to_e_reco', cuts=cuts, title_prefix='Standard Reco: ')

            make_1d_property_errors_metric_over_energy(pred_file, 'energy', (None, 'median'), main_perf_plots_path + '/1d',
                                                       'e_true_to_median_error_energy', reco_energy_correction=reco_energy_correction,
                                                       cuts=cuts, compare_2nd_reco=pred_file_2nd_reco)
            make_1d_property_errors_metric_over_energy(pred_file, 'energy', ('rel_std_div', None), main_perf_plots_path + '/1d',
                                                       'e_true_to_rel_std_div_energy', reco_energy_correction=reco_energy_correction,
                                                       cuts=cuts, compare_2nd_reco=pred_file_2nd_reco)

        if 'dir' in dataset_modifier:
            print('Generating plots for directional performance investigations')
            make_2d_prop_to_prop_plot(pred_file, 'azimuth_true', 'azimuth_reco', main_perf_plots_path + '/2d',
                                      'azimuth_true_to_azimuth_reco', cuts=cuts, title_prefix='OrcaNet: ')
            make_2d_prop_to_prop_plot(pred_file, 'zenith_true', 'zenith_reco', main_perf_plots_path + '/2d',
                                      'zenith_true_to_zenith_reco', cuts=cuts, title_prefix='OrcaNet: ')

            if pred_file_2nd_reco is not None:
                make_2d_prop_to_prop_plot(pred_file_2nd_reco, 'azimuth_true', 'azimuth_reco', main_perf_plots_path + '/2d',
                                          'standard_reco_azimuth_true_to_azimuth_reco', cuts=cuts, title_prefix='Standard Reco: ')
                make_2d_prop_to_prop_plot(pred_file_2nd_reco, 'zenith_true', 'zenith_reco', main_perf_plots_path + '/2d',
                                          'standard_reco_zenith_true_to_zenith_reco', cuts=cuts, title_prefix='Standard Reco: ')

            make_1d_property_errors_metric_over_energy(pred_file, 'dirs_vector', (None, 'median'), main_perf_plots_path + '/1d',
                                                       'e_true_to_median_error_dirs_vect', cuts=cuts,
                                                       compare_2nd_reco=pred_file_2nd_reco)
            make_1d_property_errors_metric_over_energy(pred_file, 'dirs_spherical', (None, 'median'), main_perf_plots_path + '/1d',
                                                       'e_true_to_median_error_dirs_spherical', cuts=cuts,
                                                       compare_2nd_reco=pred_file_2nd_reco)

        if 'bjorkeny' in dataset_modifier:
            print('Generating plots for bjorkeny performance investigations')
            make_2d_prop_to_prop_plot(pred_file, 'bjorkeny_true', 'bjorkeny_reco', main_perf_plots_path + '/2d',
                                      'bjorkeny_true_to_bjorkeny_reco', cuts=cuts, title_prefix='OrcaNet: ')
            if pred_file_2nd_reco is not None:
                make_2d_prop_to_prop_plot(pred_file_2nd_reco, 'bjorkeny_true', 'bjorkeny_reco', main_perf_plots_path + '/2d',
                                          'standard_reco_bjorkeny_true_to_bjorkeny_reco', cuts=cuts, title_prefix='Standard Reco: ')

            make_1d_property_errors_metric_over_energy(pred_file, 'bjorkeny', (None, 'median'), main_perf_plots_path + '/1d',
                                                       'e_true_to_median_error_bjorkeny', cuts=cuts,
                                                       compare_2nd_reco=pred_file_2nd_reco)

        if 'vtx' in dataset_modifier:
            print('Generating plots for vertex performance investigations')
            vtx_tuples = [('vtx_x_true', 'vtx_x_reco'), ('vtx_y_true', 'vtx_y_reco'), ('vtx_y_true', 'vtx_y_reco')]

            for vtx_tpl in vtx_tuples:
                make_2d_prop_to_prop_plot(pred_file, vtx_tpl[0], vtx_tpl[0], main_perf_plots_path + '/2d',
                                          vtx_tpl[0] + '_to_' + vtx_tpl[1], cuts=cuts, title_prefix='OrcaNet: ')
                if pred_file_2nd_reco is not None:
                    make_2d_prop_to_prop_plot(pred_file_2nd_reco, vtx_tpl[0], vtx_tpl[0], main_perf_plots_path + '/2d',
                                              'standard_reco_' + vtx_tpl[0] + '_to_' + vtx_tpl[1],
                                              cuts=cuts, title_prefix='Standard Reco: ')

            make_1d_property_errors_metric_over_energy(pred_file, 'vertex_vector', (None, 'median'), main_perf_plots_path + '/1d',
                                                       'e_true_to_median_error_vertex', cuts=cuts,
                                                       compare_2nd_reco=pred_file_2nd_reco)

        if 'errors' in dataset_modifier:
            print('Generating plots for error performance investigations')
            cuts = None
            make_1d_reco_err_div_by_std_dev_plot(pred_file, main_perf_plots_path + '/1d', 'reco_err_div_by_std_dev_', cuts=cuts)
            make_1d_reco_err_to_reco_residual_plot(pred_file, main_perf_plots_path + '/1d', 'reco_err_to_true_reco_residual_', cuts=cuts)
            make_2d_true_reco_plot_different_sigmas(pred_file, main_perf_plots_path + '/2d', 'true_reco_plot_different_sigmas', cuts=cuts)

    else:
        raise ValueError('The dataset_modifier ' + str(dataset_modifier) + ' is not known.')

    pred_file.close()


def make_plots_subfolders(main_perf_plots_path, dataset_modifier):
    """
    Makes the directories for the plots of a certain class_type based on the main plots dir.

    Parameters
    ----------
    main_perf_plots_path : str
        Path to the pred_performance plots folder in the OrcaNet dir structure.
    dataset_modifier : str
        String that specifies the dataset modifier.

    """
    if 'bg_classifier' in dataset_modifier:
        subfolders = ['1d']
    elif 'ts_classifier' in dataset_modifier:
        subfolders = ['1d']
    elif 'regression' in dataset_modifier:
        subfolders = ['1d', '2d']
    else:
        raise ValueError('The dataset_modifier ' + str(dataset_modifier) + ' is not known.')

    for folder in subfolders:
        if not os.path.exists(main_perf_plots_path + '/' + folder):
            print('Creating directory: ' + main_perf_plots_path + '/' + folder)
            os.makedirs(main_perf_plots_path + '/' + folder)



from unittest import TestCase
from unittest.mock import MagicMock
import os
from keras.models import Model
from keras.layers import Input, Dense
import numpy as np
import shutil

from orcanet.logging import SummaryLogger, merge_arrays, BatchLogger, gen_line_str
from orcanet.core import Organizer


class TestSummaryLogger(TestCase):
    def setUp(self):
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp")
        self.summary_file = os.path.join(self.temp_dir, "summary.txt")

        self.orga = Organizer(self.temp_dir)
        self.orga.io.get_file_sizes = MagicMock(return_value=[100, 100])
        model = build_test_model()
        self.smry = SummaryLogger(self.orga, model)

        self.metrics = model.metrics_names

    def test_writing_integration(self):
        history_train_0 = dict(zip(self.metrics, [0.0, 0.5]))
        history_train_1 = dict(zip(self.metrics, [1.0, 1.5]))
        history_val = dict(zip(self.metrics, [2.0, 2.5]))

        target = [
            "Epoch     | LR        | train_loss | val_loss  | train_mean_absolute_error | val_mean_absolute_error\n",
            "----------+-----------+------------+-----------+---------------------------+------------------------\n",
            "0.5       | 0.001     | 0          | nan       | 0.5                       | nan                    \n",
            "1         | 0.002     | 1          | nan       | 1.5                       | nan                    \n",
        ]
        filled_line = "1         | 0.002     | 1          | 2         | 1.5                       | 2.5                    \n"

        def check_file(target_lines):
            with open(self.summary_file) as file:
                for i, line in enumerate(file):
                    self.assertEqual(target_lines[i], line)

        epoch = (1, 1)
        lr = 0.001
        self.smry.write_line(self.orga.io.get_epoch_float(*epoch), lr,
                             history_train=history_train_0)
        check_file(target)

        epoch = (1, 2)
        lr = 0.002
        self.smry.write_line(self.orga.io.get_epoch_float(*epoch), lr,
                             history_train=history_train_1)
        check_file(target)

        target[-1] = filled_line
        lr = np.nan
        self.smry.write_line(self.orga.io.get_epoch_float(*epoch), lr,
                             history_val=history_val)
        check_file(target)

        check_file(target)

    def tearDown(self):
        os.remove(self.summary_file)


class TestLoggingUtil(TestCase):
    def test_merge_arrays(self):
        a = [1, 2, np.nan, np.nan, "nan", np.nan]
        b = [np.nan, 2, np.nan, 3, 4, "nan"]
        target = [1, 2, np.nan, 3, 4, "nan"]
        merged = merge_arrays(a, b)
        self.assertSequenceEqual(merged, target)

    def test_merge_arrays_error(self):
        a = [1, 1]
        b = [2, 1]
        with self.assertRaises(ValueError):
            merge_arrays(a, b)

    def test_merge_arrays_exclude(self):
        a = [1, 1]
        b = [2, 1]
        target = [1, 1]
        merged = merge_arrays(a, b, 0)
        self.assertSequenceEqual(merged, target)


class TestGenLineStr(TestCase):
    def setUp(self):
        self.widths = [9, 9]
        self.long = [1.23456789e-9, 2.0]
        self.short = ["blub", 2]

        self.kwargs = {
            "seperator": " | ",
            "float_precision": 4,
            "minimum_cell_width": 9,
        }

    def test_ok(self):
        line_l, widths_l = gen_line_str(self.long, self.widths, **self.kwargs)
        line_s, widths_s = gen_line_str(self.short, self.widths, **self.kwargs)

        #          "1---5---9_|_1---5---9"
        target_s = "blub      | 2        "
        target_l = "1.235e-09 | 2        "
        self.assertEqual(line_l, target_l)
        self.assertEqual(line_s, target_s)
        self.assertSequenceEqual(self.widths, widths_l)
        self.assertSequenceEqual(self.widths, widths_s)


class TestBatchLogger(TestCase):
    def setUp(self):
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp",
                                     "batch_logger")

        self.model = build_test_model()
        samples_file1 = 40
        samples_file2 = 60
        self.data_file1 = get_test_data(samples_file1)
        self.data_file2 = get_test_data(samples_file2)

        self.orga = Organizer(self.temp_dir)
        self.orga.io.get_file_sizes = MagicMock(return_value=[samples_file1,
                                                              samples_file2])
        self.orga.io.get_subfolder("train_log", create=True)

        self.orga.cfg.batchsize = 10
        self.orga.cfg.train_logger_display = 4

    def tearDown(self):
        """ Remove the .temp directory. """
        shutil.rmtree(self.temp_dir)

    def test_batch_logger_epoch_1_logfile_1(self):
        epoch = (1, 1)
        lines = self._make_and_get_lines(epoch)
        target_file1 = [
            'Batch       | Batch_float | loss        | mean_absolute_error\n',
            '------------+-------------+-------------+--------------------\n',
            '4           | 0.2         | 0.25        | 0.5                \n',
        ]
        for line_no in range(len(lines)):
            self.assertEqual(target_file1[line_no], lines[line_no])

    def test_batch_logger_epoch_1_logfile_2(self):
        epoch = (1, 2)
        lines = self._make_and_get_lines(epoch)
        target_file1 = [
            'Batch       | Batch_float | loss        | mean_absolute_error\n',
            '------------+-------------+-------------+--------------------\n',
            '4           | 0.6         | 0.25        | 0.5                \n',
            '6           | 0.8         | 0.125       | 0.25               \n',
        ]
        for line_no in range(len(lines)):
            self.assertEqual(target_file1[line_no], lines[line_no])

    def test_batch_logger_epoch_2_logfile_1(self):
        epoch = (2, 1)
        lines = self._make_and_get_lines(epoch)
        target_file1 = [
            'Batch       | Batch_float | loss        | mean_absolute_error\n',
            '------------+-------------+-------------+--------------------\n',
            '4           | 1.2         | 0.25        | 0.5                \n',
        ]
        for line_no in range(len(lines)):
            self.assertEqual(target_file1[line_no], lines[line_no])

    def _make_and_get_lines(self, epoch):
        """ Create some train log file with the batch logger. """
        batch_logger = BatchLogger(self.orga, epoch)

        if epoch[1] == 1:
            train_file = self.data_file1
        elif epoch[1] == 2:
            train_file = self.data_file2
        else:
            raise AssertionError(epoch)

        self.model.fit(train_file[0], train_file[1],
                       batch_size=self.orga.cfg.batchsize,
                       callbacks=[batch_logger], verbose=0)
        logfile_path = '{}/log_epoch_{}_file_{}.txt'.format(
            os.path.join(self.temp_dir, "train_log"),
            epoch[0], epoch[1])

        with open(logfile_path) as file:
            lines = file.readlines()
        return lines


def build_test_model():
    input_shape = (1,)
    inp = Input(input_shape, name="inp")
    x = Dense(1, kernel_initializer="Ones",
              bias_initializer="Zeros", trainable=False)(inp)
    outp = Dense(1, name="out", kernel_initializer="Ones",
                 bias_initializer="Zeros", trainable=False)(x)

    test_model = Model(inp, outp)
    test_model.compile("sgd", loss={"out": "mse"}, metrics={"out": "mae"})

    return test_model


def get_test_data(samples):
    """
    Make model input data.

    xs : [0.5, ]
    ys : [1, ]

    MSE: 1/4
    MAE: 1/2

    """
    dtypes = [('inp', '<f8'), ]
    xs = np.ones(samples,) * 0.5
    ys = np.ones((samples, 1)).ravel().view(dtype=dtypes)
    return xs, ys



from unittest import TestCase
from keras.layers import Input, Dense, Dropout, Activation, Convolution3D, BatchNormalization, MaxPooling3D,\
                         Convolution2D, MaxPooling2D
from keras.models import Model

from orcanet.builder_util.builders import BlockBuilder


class TestSequentialBuilder(TestCase):

    def test_input_names_and_shapes_full_model(self):
        body_defaults = {"type": "conv_block", "conv_dim": 3}
        head_defaults = {}
        builder = BlockBuilder(body_defaults, head_defaults)

        input_name = "test_input"
        input_shape = (4, 4, 4, 1)
        input_shapes = {input_name: input_shape}

        conv_layer_config = (
            {"filters": 3, },
            {"filters": 3, "pool_size": (2, 2, 2)},
            {"filters": 5, },
        )

        head_arch = "categorical"
        head_kwargs = {
            "output_name": "ts_output",
            "categories": 2,
        }

        model = builder.build(input_shapes, conv_layer_config, head_arch, head_kwargs)

        self.assertEqual([input_name, ], model.input_names)
        self.assertEqual(input_shape, model.input_shape[1:])

    def test_attach_layer_conv(self):
        inp = Input((6, 6, 1))
        body_defaults = {"type": "conv_block", "conv_dim": 2}
        layer_config = {"filters": 2, "pool_size": 2, "dropout": 0.2, "batchnorm": True, "kernel_reg": "l2"}

        builder = BlockBuilder(body_defaults, None)
        x = builder.attach_block(inp, layer_config)
        model = Model(inp, x)

        self.assertIsInstance(model.layers[1], Convolution2D)
        self.assertIsInstance(model.layers[2], BatchNormalization)
        self.assertIsInstance(model.layers[3], Activation)
        self.assertIsInstance(model.layers[4], MaxPooling2D)
        self.assertIsInstance(model.layers[5], Dropout)
        self.assertEqual(model.output_shape[1:], (3, 3, 2))

    def test_attach_output_layers_regression_output_shape_and_names(self):
        head_arch = "regression_error"
        head_arch_args = {
            "output_names": ['out_A', 'out_B'],
        }

        inp = Input((5, ))
        builder = BlockBuilder()
        x = builder.attach_output_layers(inp, head_arch, head_arch_args)
        model = Model(inp, x)

        target_output_shapes = {
            'out_A': (1,),
            'out_B': (1,),
            'out_A_err': (2,),
            'out_B_err': (2,),
        }

        output_shapes = {}
        for i, output_name in enumerate(model.output_names):
            output_shapes[output_name] = model.output_shape[i][1:]

        self.assertDictEqual(output_shapes, target_output_shapes)

    def test_attach_layer_dense(self):
        inp = Input((3, 3, 1))
        body_defaults = {"type": "conv_block", "conv_dim": 2}
        layer_config = {"type": "dense_block", "units": 5, "dropout": 0.2, "batchnorm": True}

        builder = BlockBuilder(body_defaults, None)
        x = builder.attach_block(inp, layer_config)
        model = Model(inp, x)

        self.assertIsInstance(model.layers[1], Dense)
        self.assertIsInstance(model.layers[2], BatchNormalization)
        self.assertIsInstance(model.layers[3], Activation)
        self.assertIsInstance(model.layers[4], Dropout)
        self.assertEqual(model.output_shape[1:], (3, 3, 5))

    def test_attach_layer_wrong_layer_config_conv_block(self):
        inp = Input((3, 3, 1))
        layer_config = {"type": "conv_block", "filters": 2, "units": 5}
        builder = BlockBuilder(None, None)

        with self.assertRaises(TypeError):
            builder.attach_block(inp, layer_config)

    def test_attach_layer_wrong_layer_config_dense_block(self):
        inp = Input((3, 3, 1))
        layer_config = {"type": "dense_block", "filters": 2, "units": 5}
        builder = BlockBuilder(None, None)

        with self.assertRaises(TypeError):
            builder.attach_block(inp, layer_config)




from unittest import TestCase
from unittest.mock import MagicMock, patch
from keras.models import Model
from keras.layers import Dense, Input
import os

from orcanet.core import Organizer, Configuration


class TestOrganizer(TestCase):
    def setUp(self):
        self.orga = Organizer("./")
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp", "core")

    def test_load_model_new_training(self):
        # latest epoch is None aka no model has been trained
        latest_epoch = None
        self.orga.io.get_latest_epoch = MagicMock(return_value=latest_epoch)

        # no model given = error
        target_model = None
        with self.assertRaises(ValueError):
            self.orga._get_model(target_model, logging=False)

        # model given = ok
        target_model = "the model is simply handed through"
        model = self.orga._get_model(target_model, logging=False)
        self.assertEqual(target_model, model)

    def test_load_model_continue_training(self):
        # latest epoch is not None aka model has been trained before
        latest_epoch = (1, 1)
        self.orga.io.get_latest_epoch = MagicMock(return_value=latest_epoch)

        # no model given = load model
        target_model = None
        model_file = self.temp_dir + "test_model.h5"
        self.orga.io.get_model_path = MagicMock(return_value=model_file)

        try:
            os.mkdir(self.temp_dir)
            # no model given = ok
            saved_model = build_test_model()
            saved_model.save(model_file)
            loaded_model = self.orga._get_model(model=target_model, logging=False)
            # mssing: check if models are equal (they probably are)
        finally:
            os.remove(model_file)
            os.rmdir(self.temp_dir)

        # model given = ok
        target_model = "the model is simply handed through"

        model = self.orga._get_model(target_model, logging=False)
        self.assertEqual(target_model, model)

    def test_val_is_due(self):
        no_of_files = 3
        self.orga.cfg.validate_interval = 2
        targets = {
            (1, 1): False,
            (1, 2): True,
            (1, 3): True,

            (2, 1): False,
            (2, 2): True,
        }

        self.orga.io.get_no_of_files = MagicMock(return_value=no_of_files)
        for epoch, target in targets.items():
            value = self.orga._val_is_due(epoch)
            self.assertEqual(value, target)

    @patch('orcanet.core.make_model_prediction')
    def test_predict(self, mocked):
        self.orga.cfg._list_file = "test/test.toml"
        func_args = []

        def mock_make_model_prediction(orga, model, pred_filename, samples=None):
            func_args.extend([orga, model, pred_filename, samples])

        mocked.side_effect = mock_make_model_prediction

        self.orga.load_saved_model = MagicMock(return_value=build_test_model())

        value_filepath = self.orga.predict(epoch=1, fileno=1)
        target_filepath = "./predictions/pred_model_epoch_1_file_1_on_test_val_files.h5"

        self.assertEqual(value_filepath, target_filepath)
        self.assertEqual(func_args[2], value_filepath)
        self.assertEqual(func_args[3], None)


class TestZeroCenter(TestCase):
    def setUp(self):
        self.orga = Organizer("./")
        self.orga.cfg.zero_center_folder = "nope/"

    def test_get_xs_mean_existing(self):
        target = "asd"
        self.orga.xs_mean = target
        value = self.orga.get_xs_mean(logging=False)
        self.assertEqual(target, value)

    def test_get_xs_mean_no_folder(self):
        self.orga.cfg.zero_center_folder = None
        with self.assertRaises(ValueError):
            self.orga.get_xs_mean(logging=False)

    @patch('orcanet.core.load_zero_center_data')
    def test_get_xs_mean_new(self, load_mock):
        target = "blub"

        def mock_xs_mean():
            return target

        load_mock.return_value = mock_xs_mean()
        xs_mean = self.orga.get_xs_mean(logging=False)
        self.assertEqual(target, xs_mean)
        self.assertEqual(target, self.orga.xs_mean)


class TestConfiguration(TestCase):
    def setUp(self):
        self.data_folder = os.path.join(os.path.dirname(__file__), "data")
        self.working_list = self.data_folder + "/in_out_test_list.toml"
        self.temp_dir = os.path.join(os.path.dirname(__file__), ".temp", "core")
        os.mkdir(self.temp_dir)

    def tearDown(self):
        os.rmdir(self.temp_dir)

    def test_output_folder_bracket(self):
        list_file = None
        config_file = None

        output_folder = "test"
        cfg = Configuration(output_folder, list_file, config_file)
        self.assertEqual(output_folder + "/", cfg.output_folder)

        output_folder = "test/"
        cfg = Configuration(output_folder, list_file, config_file)
        self.assertEqual(output_folder, cfg.output_folder)

    def test_listfile_already_given(self):
        config_file = None
        output_folder = "test"

        list_file = self.working_list
        cfg = Configuration(output_folder, list_file, config_file)

        with self.assertRaises(ValueError):
            cfg.import_list_file(list_file)

    def test_get_files_not_given(self):
        cfg = Configuration("./", None, None)
        with self.assertRaises(AttributeError):
            cfg.get_files("train")
        with self.assertRaises(AttributeError):
            cfg.get_files("val")
        with self.assertRaises(NameError):
            cfg.get_files("schmu")

    def test_update_config_bad_format(self):
        cfg = Configuration("./", None, None)

        content = [
            "[config]",
            "schmu=1",
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(AttributeError):
                cfg.update_config(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_ok(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            cfg.import_list_file(file_name)

        finally:
            os.remove(file_name)

    def test_load_list_file_unknown_keyword(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',
            'schmu = 1',
            
            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(ValueError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_no_train_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'schmu=1',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(NameError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_no_val_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'schmu=1',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(NameError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_different_no_of_train_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", ]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")
            with self.assertRaises(ValueError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)

    def test_load_list_file_different_no_of_val_files(self):
        cfg = Configuration("./", None, None)

        content = [
            '[input_A]',
            'train_files = ["input_A_train_1.h5", "input_A_train_2.h5"]',
            'validation_files = ["input_A_val_1.h5", "input_A_val_2.h5"]',

            '[input_B]',
            'train_files = ["input_B_train_1.h5", "input_B_train_2.h5",]',
            'validation_files = ["input_B_val_1.h5",]',
        ]
        file_name = self.temp_dir + "/test_config.toml"

        try:
            with open(file_name, "w") as f:
                for line in content:
                    f.write(line + "\n")

            with self.assertRaises(ValueError):
                cfg.import_list_file(file_name)
        finally:
            os.remove(file_name)


def build_test_model(compile=False):
    input_shape = (1,)
    inp = Input(input_shape, name="inp")
    outp = Dense(1, name="out")(inp)

    test_model = Model(inp, outp)
    if compile:
        test_model.compile("sgd", loss={"out": "mse"}, metrics={"out": "mae"})

    return test_model



from unittest import TestCase
from keras.models import Model
from keras.layers import Input, Concatenate, Dropout, Dense
import numpy as np
from keras import backend as K

from orcanet.model_builder import change_dropout_rate


class TestDropoutChange(TestCase):

    def setUp(self):
        def dropout_model(rate_before, rate_after):
            inp1 = Input((5, 1))
            x1 = Dropout(rate_before)(inp1)
            x1 = Dense(5)(x1)

            inp2 = Input((5, 1))
            x2 = Dropout(rate_before)(inp2)

            x = Concatenate(axis=-1)([x1, x2])
            x = Dense(5)(x)
            out = Dropout(rate_after)(x)

            model = Model([inp1, inp2], out)
            return model

        def get_layer_output(model, samples, layer_no=-1):
            l_out = K.function(model.input + [K.learning_phase(), ],
                               [model.layers[layer_no].output])
            # output in train mode = 1
            layer_output = l_out(samples + [1, ])[0]
            return layer_output

        def calculate_rate(model, samples, layer_no):
            layer_output = get_layer_output(model, samples, layer_no)
            rate = np.sum(layer_output == 0)/layer_output.size
            return rate

        self.calculate_rate = calculate_rate
        self.concat_layer_no = 6

        self.model_0 = dropout_model(0., 0.)
        self.xs = [np.ones((50, 5, 1)), np.ones((50, 5, 1))]

    def test_change_dropout_after_concat(self):
        model_changed = change_dropout_rate(self.model_0,
                                            before_concat=0.0,
                                            after_concat=0.999)
        rate_before_conc = self.calculate_rate(model_changed, self.xs,
                                               self.concat_layer_no)
        rate_after_conc = self.calculate_rate(model_changed, self.xs, -1)
        print("rate before Concatenate: {}\tafter: {}".format(rate_before_conc,
                                                              rate_after_conc))
        self.assertLess(rate_before_conc, 0.2)
        self.assertGreater(rate_after_conc, 0.6)

    def test_change_dropout_before_and_after_concat(self):
        model_changed = change_dropout_rate(self.model_0,
                                            before_concat=0.999,
                                            after_concat=0.999)
        rate_before_conc = self.calculate_rate(model_changed, self.xs,
                                               self.concat_layer_no)
        rate_after_conc = self.calculate_rate(model_changed, self.xs, -1)
        print("rate before Concatenate: {}\tafter: {}".format(rate_before_conc,
                                                              rate_after_conc))
        self.assertGreater(rate_before_conc, 0.6)
        self.assertGreater(rate_after_conc, 0.6)

    def test_weights_are_copied_over(self):
        model_changed = change_dropout_rate(self.model_0,
                                            before_concat=0.999,
                                            after_concat=0.999)
        for layer_no in range(len(self.model_0.layers)):
            weights_0 = self.model_0.layers[layer_no].get_weights()
            weights_changed = model_changed.layers[layer_no].get_weights()
            for i in range(len(weights_0)):
                self.assertTrue(np.array_equal(weights_0[i], weights_changed[i]))



# -*- coding: utf-8 -*-
"""
Visualization tools used with Keras.
1) Makes performance graphs for training and validating.
2) Visualizes activations for Keras models
"""
import numpy as np

import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from orcanet.utilities.nn_utilities import get_layer_output


def plot_history(
        train_data, val_data=None,
        color=None, title=None, x_label="Epoch", y_label=None, grid=True,
        legend=True, train_label="training", val_label="validation",
        x_lims=None, y_lims="auto", x_ticks=None):
    """
    Plot a training and optionally a validation line in a single plot.

    The val data can contain nan's.

    Parameters
    ----------
    train_data : List
        X data [0] and y data [1] of the train curve. Will be plotted as
        connected dots.
    val_data : List
        Optional X data [0] and y data [1] of the validation curve.
        Will be plotted as a faint solid line of the same color as train.
    color : str
        Color used for the train/val line.
    title : str
        Title of the plot.
    x_label : str
        X label of the plot.
    y_label : str
        Y label of the plot.
    grid : bool
        If true, show a grid.
    legend : bool
        If true, show a legend.
    train_label : str
        Label for the train line in the legend.
    val_label : str
        Label for the validation line in the legend.
    x_lims : List
        X limits of the data.
    y_lims : List or str
        Y limits of the data. "auto" for auto-calculation.
    x_ticks : List
        Positions of the major x ticks.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The plot.

    """
    fig, ax = plt.subplots()
    if train_data is None and val_data is None:
        raise ValueError("Can not plot when no train and val data is given.")

    if train_data is not None:
        train_plot = plt.plot(
            train_data[0], train_data[1], color=color, ls='--',
            zorder=3, label=train_label, lw=0.6, alpha=0.5)
        train_color = train_plot[0].get_color()
    else:
        train_color = color

    if val_data is not None:
        # Skip over nan values, so that all dots are connected
        not_nan = ~np.isnan(val_data[1])
        val_data = val_data[0][not_nan], val_data[1][not_nan]
        # val plot always has the same color as the train plot
        plt.plot(val_data[0], val_data[1], color=train_color,
                 marker='o', zorder=3, label=val_label)

    if x_ticks is None:
        x_ticks = get_epoch_xticks(train_data, val_data)
    plt.xticks(x_ticks)

    if x_lims is not None:
        plt.xlim(x_lims)

    if y_lims is not None:
        if y_lims == "auto":
            y_lims = get_ylims(train_data, val_data, fraction=0.25)
        plt.ylim(y_lims)

    if legend:
        ax.legend(loc='upper right')
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    if title is not None:
        title = plt.title(title)
        title.set_position([.5, 1.04])
    if grid:
        plt.grid(True, zorder=0, linestyle='dotted')

    return fig


def get_ylims(train_data, val_data=None, fraction=0.25):
    """
    Get the y limits for the summary plot.

    For the training data, limits are calculated while ignoring data points
    which are far from the median (in terms of the median distance
    from the median).
    This is because there are outliers sometimes in the training data,
    especially early on in the training.

    Parameters
    ----------
    train_data : List
        X data (0) and y data (1) of the train curve. Will be plotted as
        connected dots.
    val_data : List or None
        X data (0) and y data (1) of the validation curve. Will be plotted
        as a faint solid line.
    fraction : float
        How much whitespace of the total y range is added above and below
        the lines.

    Returns
    -------
    y_lims : tuple
        Minimum, maximum of the data.

    """
    assert not (train_data is None and
                val_data is None), "train and val data are None"

    def reject_outliers(data, threshold):
        d = np.abs(data - np.median(data))
        mdev = np.median(d)
        s = d / mdev if mdev else 0.
        no_outliers = data[s < threshold]
        lims = np.amin(no_outliers), np.amax(no_outliers)
        return lims

    mins, maxs = [], []
    if train_data is not None:
        y_train = train_data[1]
        y_lims_train = reject_outliers(y_train, 5)
        mins.append(y_lims_train[0])
        maxs.append(y_lims_train[1])

    if val_data is not None:
        y_val = val_data[1]

        if len(y_val) == 1:
            y_lim_val = y_val[0], y_val[0]
        else:
            y_lim_val = np.amin(y_val), np.amax(y_val)

        mins.append(y_lim_val[0])
        maxs.append(y_lim_val[1])

    if len(mins) == 1:
        y_lims = (mins[0], maxs[0])
    else:
        y_lims = np.amin(mins), np.amax(maxs)

    if y_lims[0] == y_lims[1]:
        y_range = 0.1 * y_lims[0]
    else:
        y_range = y_lims[1] - y_lims[0]

    if fraction != 0:
        y_lims = (y_lims[0] - fraction * y_range,  y_lims[1] + fraction * y_range)

    return y_lims


def get_epoch_xticks(train_data, val_data=None):
    """
    Calculates the xticks for the train and validation summary plot.

    One tick per poch. Less the larger #epochs is.

    Parameters
    ----------
    train_data : List
        X data (0) and y data (1) of the train curve. Will be plotted as
        connected dots.
    val_data : List or None
        X data (0) and y data (1) of the validation curve. Will be plotted
        as a faint solid line.

    Returns
    -------
    x_ticks_major : numpy.ndarray
        Array containing the ticks.

    """
    assert not (train_data is None and
                val_data is None), "train and val data are None"
    x_points = np.array([])

    if train_data is not None:
        x_points = np.append(x_points, train_data[0])
    if val_data is not None:
        x_points = np.append(x_points, val_data[0])

    minimum, maximum = np.amin(x_points), np.amax(x_points)
    start_epoch, end_epoch = np.floor(minimum), np.ceil(maximum)
    # reduce number of x_ticks by factor of 2 if n_epochs > 20
    n_epochs = end_epoch - start_epoch
    x_ticks_stepsize = 1 + np.floor(n_epochs / 20.)
    x_ticks_major = np.arange(
        start_epoch, end_epoch + x_ticks_stepsize, x_ticks_stepsize)

    return x_ticks_major


def update_summary_plot(orga):
    """
    Plot and save all metrics of the given validation- and train-data
    into a pdf file, each metric in its own plot.

    If metric pairs of a variable and its error are found (e.g. e_loss
    and e_err_loss), they will have the same color and appear back to
    back in the plot.

    Parameters
    ----------
    orga : object Organizer
        Contains all the configurable options in the OrcaNet scripts.

    """
    plt.ioff()
    pdf_name = orga.io.get_subfolder("plots", create=True) + "/summary_plot.pdf"

    # Extract the names of the metrics
    all_metrics = orga.history.get_metrics()
    # Sort them
    all_metrics = sort_metrics(all_metrics)
    # Plot them w/ custom color cycle
    colors = ['#000000', '#332288', '#88CCEE', '#44AA99', '#117733', '#999933',
              '#DDCC77', '#CC6677', '#882255', '#AA4499', '#661100', '#6699CC',
              '#AA4466', '#4477AA']  # ref. personal.sron.nl/~pault/
    color_counter = 0
    with PdfPages(pdf_name) as pdf:
        for metric_no, metric in enumerate(all_metrics):
            # If this metric is an err metric of a variable, color it the same
            if all_metrics[metric_no-1] == metric.replace("_err", ""):
                color_counter -= 1
            fig = orga.history.plot_metric(
                metric, color=colors[color_counter % len(colors)])
            color_counter += 1
            pdf.savefig(fig)
            plt.close(fig)

        lr_fig = orga.history.plot_lr()
        color_counter += 1
        pdf.savefig(lr_fig)
        plt.close(lr_fig)


def sort_metrics(metric_names):
    """
    Sort a list of metrics, so that errors are right after their variable.
    The format of the metric names have to be e.g. e_loss and e_err_loss
    for this to work.

    Example
    ----------
    >>> sort_metrics( ['e_loss', 'loss', 'e_err_loss', 'dx_err_loss'] )
    ['e_loss', 'e_err_loss', 'loss', 'dx_err_loss']

    Parameters
    ----------
    metric_names : List
        List of metric names.

    Returns
    -------
    sorted_metrics : List
        List of sorted metric names with the same length as the input.

    """
    sorted_metrics = [0] * len(metric_names)
    counter = 0
    for metric_name in metric_names:
        if "err_" in metric_name:
            if metric_name.replace("err_", "") not in metric_names:
                sorted_metrics[counter] = metric_name
                counter += 1
            continue
        sorted_metrics[counter] = metric_name
        counter += 1
        err_loss = metric_name.split("_loss")[0]+"_err_loss"
        if err_loss in metric_names:
            sorted_metrics[counter] = err_loss
            counter += 1

    assert 0 not in sorted_metrics, "Something went wrong with the sorting of " \
                                    "metrics! Given was {}, output was " \
                                    "{}. ".format(metric_names, sorted_metrics)

    return sorted_metrics


def plot_activations(model, samples, layer_name, mode='test', bins=100):
    """
    Make plots of activations of one layer of a model.

    Arrays will be flattend before plotting them as histograms.

    Parameters
    ----------
    model : keras model
        The model to make the data with.
    samples : dict
        Input data.
    layer_name : str or None
        Name of the layer to get info from. None for all layers.
    mode : str
        Mode of the layers during the forward pass. Either train or test.
        Important for batchnorm, dropout, ...
    bins : int
        Number of bins of the histogram.

    Returns
    -------
    fig : plt figure
        The plot of the activations of the given layer.

    """

    layer = model.get_layer(layer_name)
    if layer.name in model.input_names:
        activations = samples[layer.name]
    else:
        activations = get_layer_output(model, samples, layer.name, mode)

    fig, ax = plt.subplots()
    ax.hist(activations.flatten(), bins=bins)
    ax.set_title('Activations for layer ' + str(layer_name))
    ax.set_xlabel('Activation (layer output)')
    ax.set_ylabel('Quantity [#]')

    return fig


def plot_weights(model, layer_name, bins=100):
    """
    Make plots of the weights of one layer of a model.

    Arrays will be flattend before plotting them as histograms.

    Parameters
    ----------
    model : keras model
        The model to make the data with.
    layer_name : str or None
        Name of the layer to get info from. None for all layers.
    bins : int
        Number of bins of the histogram.

    Returns
    -------
    fig : plt figure or None
        The plot of the weights of the given layer. None if the layer
        has no weights.

    """
    layer = model.get_layer(layer_name)
    layer_weights = layer.get_weights()

    if not layer_weights:
        return None

    fig, ax = plt.subplots()

    # layer_weights is a list of np arrays; flatten it
    weights = np.array([])
    for j, w_temp in enumerate(layer_weights):
        w_temp_flattened = np.array(w_temp.flatten(), dtype=np.float64)
        weights = np.concatenate((weights, w_temp_flattened), axis=0)

    ax.hist(weights, bins=bins)
    ax.set_title('Weights for layer ' + str(layer_name))
    ax.set_xlabel('Weight')
    ax.set_ylabel('Quantity [#]')
    fig.tight_layout()

    return fig



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utility code for making nn reco performance plots.
"""

import numpy as np


def select_track_shower(ptype, is_cc):
    """
    Returns a boolean mask which specifies, which events in the rows of the particle_type and is_cc arrays belong to
    the track and which to the shower class.

    Parameters
    ----------
    ptype : ndarray(ndim=1)
        Array with particle_types of some events.
    is_cc : ndarray(ndim=1)
        Array with is_cc of some events.

    Returns
    -------
    is_track : ndarray(ndim=1)
        Boolean array which specifies, if the events from the ptype & is_cc arrays belong to the track class or not.
    is_shower : ndarray(ndim=1)
        Boolean array which specifies, if the events from the ptype & is_cc arrays belong to the shower class or not.

    """
    abs_particle_type = np.abs(ptype)
    is_track = np.logical_and(abs_particle_type == 14, is_cc is True)
    is_shower = np.logical_or(abs_particle_type == 16, abs_particle_type == 12)

    return is_track, is_shower


def select_ic(ptype, is_cc, interaction_channel):
    """
    Returns a boolean mask which specifies, which events in the rows of the particle_type and is_cc arrays belong to
    the specified interaction_channel.

    Parameters
    ----------
    ptype : ndarray(ndim=1)
        Array with particle_types of some events.
    is_cc : ndarray(ndim=1)
        Array with is_cc of some events.
    interaction_channel : str
        String, that specifies the interaction channel (one of "muon-CC", "elec-CC", "elec-NC", "tau-CC")

    Returns
    -------
    is_ic : ndarray(ndim=1)
        Boolean mask array which specifies if each event belongs to the input ic or not.

    """
    ic_dict = {'muon-CC': (14, 1), 'elec-CC': (12, 1), 'elec-NC': (12, 0), 'tau-CC': (16, 1)}
    if interaction_channel not in list(ic_dict.keys()):
        raise ValueError('The interaction_channel ' + str(interaction_channel) + ' is not known.')

    abs_ptype = np.abs(ptype)
    is_ic = np.logical_and(abs_ptype == ic_dict[interaction_channel][0],
                           is_cc == ic_dict[interaction_channel][1])

    return is_ic


def correct_reco_energy(mc_info, energy_pred_array, metric='median'):
    """
    Makes the correction factors based on e-CC events, applies them to ALL shower events (e-CC, e-NC, tau-CC) and
    returns the corrected pred_energy array.

    Parameters
    ----------
    mc_info : h5py.dataset.Dataset/ndarray(ndim=2)
        The mc_info structured array of an OrcaNet nn prediction file.
    energy_pred_array : ndarray(ndim=1)
        Array, which contains the predicted energies of a nn.
    metric : str
        Metric that should be used for the correction. Available: 'median' and 'mean'.

    Returns
    -------
    energy_pred : ndarray(ndim=1)
        Array containing the corrected prediction energies.

    """
    particle_type, is_cc = mc_info['particle_type'], mc_info['is_cc']

    is_track, is_shower = select_track_shower(particle_type, is_cc)
    is_ic = select_ic(particle_type, is_cc, 'elec-CC')

    energy_mc, energy_pred = mc_info['energy'][is_ic], np.copy(energy_pred_array[is_ic])  # TODO test if this works or is actually necessary

    correction_factors_x, correction_factors_y = [], []

    e_range = np.logspace(np.log(3)/np.log(2), np.log(100)/np.log(2), 50, base=2)
    n_ranges = e_range.shape[0] - 1
    for i in range(n_ranges):
        e_range_low, e_range_high = e_range[i], e_range[i+1]
        e_range_mean = (e_range_high + e_range_low) / float(2)

        e_mc_cut_boolean = np.logical_and(e_range_low < energy_mc, energy_mc <= e_range_high)
        e_mc_cut = energy_mc[e_mc_cut_boolean]
        e_pred_cut = energy_pred[e_mc_cut_boolean]

        if metric == 'median':
            correction_factor = np.median((e_pred_cut - e_mc_cut) / e_pred_cut)
        elif metric == 'mean':
            correction_factor = np.mean((e_pred_cut - e_mc_cut) / e_pred_cut)
        else:
            raise ValueError('The specified metric "' + metric + '" is not implemented.')

        correction_factors_x.append(e_range_mean)
        correction_factors_y.append(correction_factor)

    # linear interpolation of correction factors
    energy_pred_orig_shower = energy_pred_array[is_shower]
    correction_factor_en_pred = np.interp(energy_pred_orig_shower, correction_factors_x, correction_factors_y)

    # apply correction to ALL shower ic's (including all taus atm)
    energy_pred[is_shower] = energy_pred_orig_shower + (- correction_factor_en_pred) * energy_pred_orig_shower

    return energy_pred


# --------------------------- Code for making cuts --------------------------- #

def get_event_selection_mask(mc_info, invert=False, cut_name='1-5GeV_and_3-100GeV_prod'):
    """
    Function that checks, which events in the mc_info input array are also in the event selection file,
    specified by the cut_name parameter.

    Parameters
    ----------
    mc_info : ndarray(ndim=2)
        The mc_info structured array of an OrcaNet nn prediction file.
    invert : Instead of selecting all events that survive some cuts, this _removes_ all the events that are normally
             cut and only leaves the events that are not contained in the cut_file specified by the cut_name.
    cut_name : str
        String, which specifies the actual cut file that should be loaded.
    Returns
    -------
    bool_evt_selected : ndarray(ndim=1)
        Boolean array, which specifies which events in the mc_info are contained in the cut_file specified by
        the cut_name parameter.

    """
    # load array with run_id, event_id, prod_id, particle_type and is_cc info (in that column order!)
    arr_sel_events = load_event_selection_file(cut_name)

    # select the same information from the input mc_info
    ax = np.newaxis()
    mc_info_necessary_info = np.concatenate([mc_info['run_id'][:, ax], mc_info['event_id'][:, ax],
                                             mc_info['prod_id'][:, ax], mc_info['particle_type'][:, ax],
                                             mc_info['is_cc'][:, ax]], axis=1)

    bool_evt_selected = in_nd(mc_info_necessary_info, arr_sel_events)

    if invert is True:
        bool_evt_selected = np.invert(bool_evt_selected)

    return bool_evt_selected


def load_event_selection_file(cut_name, dirpath='/home/saturn/capn/mppi033h/Data/event_selections'):
    """
    Loads a .npy file base on the cut_name and the dirpath, which contains the event selection info,
    e.g. of some other reco.

    Parameters
    ----------
    cut_name : str
        String, which specifies the actual cut file that should be loaded.
    dirpath : str
        Path to the directory, where the file specified by the cut_name is located.

    Returns
    -------
    selected_events_array : ndarray(ndim=2)
        Array with the run_id, event_id, prod_id, particle_type and is_cc info of the selected events.

    """
    if cut_name == 'neutrino_regr':
        # contains run_id, event_id, prod_id, particle_type and is_cc
        selected_events_array = np.load(dirpath + '/evt_selection_regression.npy')

    elif cut_name == 'neutrino_ts':
        # contains run_id, event_id, prod_id, particle_type and is_cc
        selected_events_array = np.load(dirpath + '/evt_selection_ts_classifier.npy')

    elif cut_name == 'bg_classifier':
        # contains run_id, event_id, prod_id, particle_type and is_cc
        selected_events_array = np.load(dirpath + '/evt_selection_bg_classifier.npy')

    else:
        raise ValueError('The specified cut_name "' + str(cut_name) + '" is not available.')

    return selected_events_array


def asvoid(arr):
    """
    Based on http://stackoverflow.com/a/16973510/190597 (Jaime, 2013-06)
    View the array as dtype np.void (bytes). The items along the last axis are
    viewed as one value. This allows comparisons to be performed on the entire row.
    """
    arr = np.ascontiguousarray(arr)
    if np.issubdtype(arr.dtype, np.floating):
        """ Care needs to be taken here since
        np.array([-0.]).view(np.void) != np.array([0.]).view(np.void)
        Adding 0. converts -0. to 0.
        """
        arr += 0.
    return arr.view(np.dtype((np.void, arr.dtype.itemsize * arr.shape[-1])))


def in_nd(a, b, assume_unique=False):
    """
    Function that generalizes the numpy in_1d function to nd. The docs are specifically for the 2d case though.

    Checks if entries in axis_0 of a exist in b and returns the bool array for all rows.

    Parameters
    ----------
    a : ndarray(ndim=2)
        Array where it should be checked whether each row exists in b or not.
    b : ndarray(ndim>=2)
        Array upon which the rows of a are checked.
    assume_unique : bool
        If True, the input arrays are both assumed to be unique, which can speed up the calculation.

    Returns
    -------
     a_in_b: ndarray(ndim=1)
        Boolean array that specifies for each row of a if it also exists in b or not.

    """
    a = asvoid(a)
    b = asvoid(b)
    a_in_b = np.in1d(a, b, assume_unique)
    return a_in_b


def test_in_nd():
    a = np.array([[1, 129385, 1, -1, -0], [1, 1, 0, 0, 0], [0, 0, 0, 0, 0]])
    b = np.array([[1, 1, 0, 0, 2], [6, 6, 6, 6, 6], [1, 129385, 1, -1, -0]])

    c = in_nd(a, b)
    it_works = np.all(c == [True, False, False])
    return it_works



# --------------------------- Code for making cuts --------------------------- #



from unittest import TestCase
import numpy as np

from orcanet.utilities.visualization import get_ylims, get_epoch_xticks, sort_metrics


class TestFunctions(TestCase):
    def setUp(self):
        train_epoch_data = np.linspace(0, 2, num=100)
        train_y_data = np.linspace(1, 2, num=100)

        val_epoch_data = np.linspace(0, 2, num=6)
        val_y_data = np.linspace(2, 3, num=6)

        self.train_data = (train_epoch_data, train_y_data)
        self.val_data = (val_epoch_data, val_y_data)

    def test_get_ylims_frac_25(self):
        y_lims = get_ylims(self.train_data, self.val_data)
        target = (0.5, 3.5)
        self.assertSequenceEqual(y_lims, target)

    def test_get_ylims_frac_0(self):
        y_lims = get_ylims(self.train_data, self.val_data, fraction=0)
        target = (1, 3)
        self.assertSequenceEqual(y_lims, target)

    def test_get_ylims_no_val(self):
        y_lims = get_ylims(self.train_data, None)
        target = (0.75, 2.25)
        self.assertSequenceEqual(y_lims, target)

    def test_get_ylims_no_train(self):
        y_lims = get_ylims(None, self.val_data)
        target = (1.75, 3.25)
        self.assertSequenceEqual(y_lims, target)

    def test_get_ylims_one_point_val(self):
        val_data = ([1., ], [3., ])
        y_lims = get_ylims(None, val_data)
        target = (2.925, 3.075)
        self.assertSequenceEqual(y_lims, target)

    def test_get_epoch_xticks(self):
        ticks = get_epoch_xticks(self.train_data, self.val_data)
        target = np.array([0, 1, 2])
        np.testing.assert_array_equal(ticks, target)

    def test_get_epoch_xticks_large(self):
        train_data = ([0.9, 21.9], [1, 2])
        ticks = get_epoch_xticks(train_data, None)
        target = np.arange(0, 23, 2)
        np.testing.assert_array_equal(ticks, target)

    def test_sort_metrics(self):
        metrics = ['e_loss', 'loss', 'e_err_loss', 'dx_err_loss']
        value = sort_metrics(metrics)
        target = ['e_loss', 'e_err_loss', 'loss', 'dx_err_loss']
        self.assertSequenceEqual(value, target)

    def test_sort_metrics_err_first(self):
        metrics = ['e_err_loss', 'dx_err_loss', 'e_loss', 'loss', ]
        value = sort_metrics(metrics)
        target = ['dx_err_loss', 'e_loss', 'e_err_loss', 'loss']
        self.assertSequenceEqual(value, target)

    def test_sort_metrics_no_err(self):
        metrics = ['e_loss', 'loss', 'blub']
        value = sort_metrics(metrics)
        target = ['e_loss', 'loss', 'blub']
        self.assertSequenceEqual(value, target)

    def test_sort_metrics_only_err(self):
        metrics = ['e_err_loss', 'dx_err_loss']
        value = sort_metrics(metrics)
        target = ['e_err_loss', 'dx_err_loss']
        self.assertSequenceEqual(value, target)




#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Utility code regarding reading user input, and writing output like logfiles.
"""

import os
import shutil
import h5py
import numpy as np

from orcanet.utilities.nn_utilities import get_inputs
from orcanet.utilities.visualization import plot_history


class IOHandler(object):
    """
    Access info indirectly contained in the cfg object.
    """
    def __init__(self, cfg):
        self.cfg = cfg

        self._used_ssd = False
        self._tmpdir_train_files = None
        self._tmpdir_val_files = None

    def get_latest_epoch(self, epoch=-1):
        """
        Check all saved models in the ./saved_models folder and return the
        highest epoch / file_no pair.

        Will only consider files that end with .h5 as models.

        Parameters
        ----------
        epoch : int
            Return the highest filenumber for this epoch. -1 for the most
            recent epoch.

        Returns
        -------
        latest_epoch : tuple or None
            The highest epoch, file_no pair. None if the folder is
            empty or does not exist yet.

        """
        saved_models_folder = self.cfg.output_folder + "saved_models"
        if os.path.exists(saved_models_folder):
            files = []
            for file in os.listdir(saved_models_folder):
                if file.endswith('.h5'):
                    files.append(file)

            if len(files) == 0:
                latest_epoch = None
            else:
                epochs = []
                for file in files:
                    # model_epoch_XX_file_YY
                    file_base = os.path.splitext(file)[0]
                    f_epoch, file_no = file_base.split(
                        "model_epoch_")[-1].split("_file_")

                    if epoch == -1 or int(f_epoch) == epoch:
                        epochs.append((int(f_epoch), int(file_no)))

                if len(epochs) == 0:
                    raise ValueError("Can not get most recent file for "
                                     "epoch {}: No files found!".format(epoch))
                latest_epoch = max(epochs)
        else:
            latest_epoch = None

        if epoch != -1 and latest_epoch is None:
            raise ValueError("Can not get most recent file for "
                             "epoch {}: No files found!".format(epoch))

        return latest_epoch

    def get_next_epoch(self, epoch):
        """
        Return the next epoch / fileno tuple.

        It depends on how many train files there are.

        Parameters
        ----------
        epoch : tuple or None
            Current epoch and file number.

        Returns
        -------
        next_epoch : tuple
            Next epoch and file number.

        """
        if epoch is None:
            next_epoch = (1, 1)
        elif epoch[1] == self.get_no_of_files("train"):
            next_epoch = (epoch[0] + 1, 1)
        else:
            next_epoch = (epoch[0], epoch[1] + 1)
        return next_epoch

    def get_previous_epoch(self, epoch):
        """ Return the previous epoch / fileno tuple. """
        if epoch[1] == 1:
            if epoch[0] == 1:
                raise ValueError("Can not get previous epoch of epoch {} file {}".format(*epoch))
            n_train_files = self.get_no_of_files("train")
            prev_epoch = (epoch[0]-1, n_train_files)

        else:
            prev_epoch = (epoch[0], epoch[1]-1)

        return prev_epoch

    def get_subfolder(self, name=None, create=False):
        """
        Get the path to one or all subfolders of the main folder.

        Parameters
        ----------
        name : str or None
            The name of the subfolder.
        create : bool
            If the subfolder should be created if it does not exist.

        Returns
        -------
        subfolder : str or tuple
            The path of the subfolder. If name is None, all subfolders
            will be returned as a tuple.

        """
        subfolders = {
            "train_log": self.cfg.output_folder + "train_log",
            "saved_models": self.cfg.output_folder + "saved_models",
            "plots": self.cfg.output_folder + "plots",
            "activations": self.cfg.output_folder + "plots/activations",
            "predictions": self.cfg.output_folder + "predictions",
        }

        def get(fdr):
            subfdr = subfolders[fdr]
            if create and not os.path.exists(subfdr):
                print("Creating directory: " + subfdr)
                os.makedirs(subfdr)
            return subfdr

        if name is None:
            subfolder = [get(name) for name in subfolders]
        else:
            subfolder = get(name)
        return subfolder

    def get_model_path(self, epoch, fileno):
        """ Get the path to a model (which might not exist yet). """
        if epoch == -1 and fileno == -1:
            epoch, fileno = self.get_latest_epoch()
        if epoch < 1 or fileno < 1:
            raise ValueError("Invalid epoch/file number {}, {}: Must be "
                             "either (-1, -1) or both >1".format(epoch, fileno))

        model_filename = self.get_subfolder("saved_models") \
            + '/model_epoch_{}_file_{}.h5'.format(epoch, fileno)
        return model_filename

    def get_pred_path(self, epoch, fileno):
        """ Get the path to a saved prediction. """
        list_file = self.cfg.get_list_file()
        if list_file is None:
            raise ValueError("No tom list file specified. Can not look up "
                             "saved prediction")
        list_name = os.path.splitext(os.path.basename(list_file))[0]

        pred_filename = self.get_subfolder("predictions") + \
            '/pred_model_epoch_{}_file_{}_on_{}_val_files.h5'.format(
                epoch, fileno, list_name)

        return pred_filename

    def use_local_node(self):
        """
        Copies the test and val files to the node-local ssd scratch folder
        and sets the new filepaths of the train and val data.
        Speeds up I/O and reduces RRZE network load.
        """
        if not self._used_ssd:
            train_files_ssd, val_files_ssd = use_local_tmpdir(
                self.cfg.get_files("train"), self.cfg.get_files("val"))
            self._tmpdir_train_files = train_files_ssd
            self._tmpdir_val_files = val_files_ssd
            self._used_ssd = True

    def get_local_files(self, which):
        """
        Get the training or validation file paths for each list input set.

        Returns the path to the copy of the file on the local tmpdir, if
        it has been made.

        Returns
        -------
        dict
            A dict containing the paths to the training or validation files on
            which the model will be trained on. Example for the format for
            two input sets with two files each:
            {
             "input_A" : ('path/to/set_A_file_1.h5', 'path/to/set_A_file_2.h5'),
             "input_B" : ('path/to/set_B_file_1.h5', 'path/to/set_B_file_2.h5'),
            }

        """
        if which == "train":
            if not self._used_ssd:
                return self.cfg.get_files("train")
            else:
                return self._tmpdir_train_files

        elif which == "val":
            if not self._used_ssd:
                return self.cfg.get_files("val")
            else:
                return self._tmpdir_val_files
        else:
            raise NameError("Unknown fileset name ", which)

    def get_n_bins(self):
        """
        Get the number of bins from the training files.

        Only the first files are looked up, the others should be identical.

        Returns
        -------
        n_bins : dict
            Toml-list input names as keys, list of the bins as values.

        """
        # TODO check if bins are equal in all files?
        train_files = self.get_local_files("train")
        n_bins = {}
        for input_key in train_files:
            with h5py.File(train_files[input_key][0], "r") as f:
                n_bins[input_key] = f[self.cfg.key_samples].shape[1:]
        return n_bins

    def get_file_sizes(self, which):
        """
        Get the number of samples in each training or validation input file.

        Parameters
        ----------
        which : str
            Either train or val.

        Returns
        -------
        file_sizes : List
            Its length is equal to the number of files in each input set.

        Raises
        ------
        ValueError
            If there is a different number of samples in any of the
            files of all inputs.

        """
        file_sizes_full, error_file_sizes, file_sizes = {}, [], []
        for n, file_no_set in enumerate(self.yield_files(which)):
            # the number of samples in the n-th file of all inputs
            file_sizes_full[n] = [h5_get_number_of_rows(
                file, datasets=[self.cfg.key_labels, self.cfg.key_samples])
                for file in file_no_set.values()]
            if not file_sizes_full[n].count(file_sizes_full[n][0]) == \
                    len(file_sizes_full[n]):
                    error_file_sizes.append(n)
            else:
                file_sizes.append(file_sizes_full[n][0])

        if len(error_file_sizes) != 0:
            err_msg = "The files you gave for the different inputs of the model " \
                      "do not all have the same number of samples!\n"
            for n in error_file_sizes:
                err_msg += "File no {} in {} has the following files sizes " \
                           "for the different inputs: {}\n".format(
                            n, which, file_sizes_full[n])
            raise ValueError(err_msg)

        return file_sizes

    def get_no_of_files(self, which):
        """
        Return the number of training or validation files.

        Only looks up the no of files of one (random) list input, as equal
        length is checked during read in.

        Parameters
        ----------
        which : str
            Either train or val.

        Returns
        -------
        no_of_files : int
            The number of files.

        """
        files = self.get_local_files(which)
        no_of_files = len(list(files.values())[0])
        return no_of_files

    def yield_files(self, which):
        """
        Yield a training or validation filepaths for every input.

        They will be yielded in the same order as they are given in the
        toml file.

        Parameters
        ----------
        which : str
            Either train or val.

        Yields
        ------
        files_dict : dict
            Keys: The name of every toml list input.
            Values: One of the filepaths.

        """
        files = self.get_local_files(which)
        for file_no in range(self.get_no_of_files(which)):
            files_dict = {key: files[key][file_no] for key in files}
            yield files_dict

    def get_file(self, which, file_no):
        """ Get a dict with the n-th files. """
        files = self.get_local_files(which)
        files_dict = {key: files[key][file_no-1] for key in files}
        return files_dict

    def check_connections(self, model):
        """
        Check if the names and shapes of the samples and labels in the
        given input files work with the model.

        Also takes into account the possibly present sample or label modifiers.

        Parameters
        ----------
        model : ks.model
            A keras model.

        Raises
        ------
        ValueError
            If they dont work together.

        """
        print("\nInput check\n-----------")
        # Get a batch of data to investigate the given modifier functions
        xs, y_values = self.get_batch()
        layer_inputs = get_inputs(model)
        # keys: name of layers, values: shape of input
        layer_inp_shapes = {key: layer_inputs[key].input_shape[1:]
                            for key in layer_inputs}
        list_inp_shapes = self.get_n_bins()

        print("The inputs in your toml list file have the following "
              "names and shapes:")
        for list_key in list_inp_shapes:
            print("\t{}\t{}".format(list_key, list_inp_shapes[list_key]))

        if self.cfg.sample_modifier is None:
            print("\nYou did not specify a sample modifier.")
        else:
            modified_xs = self.cfg.sample_modifier(xs)
            modified_shapes = {modi_key: modified_xs[modi_key].shape[1:]
                               for modi_key in modified_xs}
            print("\nAfter applying your sample modifier, they have the "
                  "following names and shapes:")
            for list_key in modified_shapes:
                print("\t{}\t{}".format(list_key, modified_shapes[list_key]))
            list_inp_shapes = modified_shapes

        print("\nYour model requires the following input names and shapes:")
        for layer_key in layer_inp_shapes:
            print("\t{}\t{}".format(layer_key, layer_inp_shapes[layer_key]))

        # Both inputs are dicts with  name: shape  of input/output layers/data
        err_inp_names, err_inp_shapes = [], []
        for layer_name in layer_inp_shapes:
            if layer_name not in list_inp_shapes.keys():
                # no matching name
                err_inp_names.append(layer_name)
            elif list_inp_shapes[layer_name] != layer_inp_shapes[layer_name]:
                # no matching shape
                err_inp_shapes.append(layer_name)

        err_msg_inp = ""
        if len(err_inp_names) == 0 and len(err_inp_shapes) == 0:
            print("\nInput check passed.")
        else:
            print("\nInput check failed!")
            if len(err_inp_names) != 0:
                err_msg_inp += "No matching input name from the list file " \
                               "for input layer(s): " + (
                                ", ".join(str(e) for e in err_inp_names) + "\n")
            if len(err_inp_shapes) != 0:
                err_msg_inp += "Shapes of layers and labels do not match for " \
                               "the following input layer(s): " + (
                                ", ".join(str(e) for e in err_inp_shapes) + "\n")
            print("Error:", err_msg_inp)

        # ----------------------------------
        print("\nOutput check\n------------")
        # tuple of strings
        mc_names = y_values.dtype.names
        print("The following {} label names are in your toml list file:".format(
            len(mc_names)))
        print("\t" + ", ".join(str(name) for name in mc_names), end="\n\n")

        if self.cfg.label_modifier is not None:
            label_names = tuple(self.cfg.label_modifier(y_values).keys())
            print("The following {} labels get produced from them by your "
                  "label_modifier:".format(len(label_names)))
            print("\t" + ", ".join(str(name) for name in label_names), end="\n\n")
        else:
            label_names = mc_names
            print("You did not specify a label_modifier. The output layers "
                  "will be provided with labels that match their name from "
                  "the above.\n\n")

        # tuple of strings
        loss_names = tuple(model.output_names)
        print("Your model has the following {} output layers:".format(
            len(loss_names)))
        print("\t" + ", ".join(str(name) for name in loss_names), end="\n\n")

        err_out_names = []
        for loss_name in loss_names:
            if loss_name not in label_names:
                err_out_names.append(loss_name)

        err_msg_out = ""
        if len(err_out_names) == 0:
            print("Output check passed.\n")
        else:
            print("Output check failed!")
            if len(err_out_names) != 0:
                err_msg_out += "No matching label name from the list file " \
                               "for output layer(s): " + (
                                ", ".join(str(e) for e in err_out_names) + "\n")
            print("Error:", err_msg_out)

        err_msg = err_msg_inp + err_msg_out
        if err_msg != "":
            raise ValueError(err_msg)

    def get_batch(self):
        """ For testing purposes, return a batch of samples and mc_infos. """
        # TODO gets mc_info only from first train file
        files_dict = next(self.yield_files("train"))
        xs = {}
        for i, inp_name in enumerate(files_dict):
            with h5py.File(files_dict[inp_name], "r") as f:
                xs[inp_name] = f[self.cfg.key_samples][:self.cfg.batchsize]
                if i == 0:
                    mc_info = f[self.cfg.key_labels][:self.cfg.batchsize]
        return xs, mc_info

    def get_input_shapes(self):
        """
        Get the input names and shapes of the data after the modifier has
        been applied.

        Returns
        -------
        input_shapes : dict
            Keys: Name of the inputs of the model.
            Values: Their shape without the batchsize.

        """
        if self.cfg.sample_modifier is None:
            input_shapes = self.get_n_bins()
        else:
            xs, mc_info = self.get_batch()
            xs_mod = self.cfg.sample_modifier(xs)
            input_shapes = {input_name: input_xs.shape[1:]
                            for input_name, input_xs in xs_mod.items()}
        return input_shapes

    def print_log(self, lines, logging=True):
        """ Print and also log to the full log file. """
        if isinstance(lines, str):
            lines = [lines, ]

        if not logging:
            for line in lines:
                print(line)
        else:
            full_log_file = self.cfg.output_folder + 'log.txt'
            with open(full_log_file, 'a+') as f_out:
                for line in lines:
                    f_out.write(line + "\n")
                    print(line)

    def get_epoch_float(self, epoch, fileno):
        """ Make a float value out of epoch/fileno. """
        # calculate the fraction of samples per file compared to all files,
        # e.g. [100, 50, 50] --> [0.5, 0.75, 1]
        file_sizes = self.get_file_sizes("train")
        file_sizes_rltv = np.cumsum(file_sizes) / np.sum(file_sizes)

        epoch_float = epoch - 1 + file_sizes_rltv[fileno - 1]
        return epoch_float


class HistoryHandler:
    """
    For reading and plotting data from summary and train log files.
    """
    def __init__(self, summary_file, train_log_folder):
        self.summary_filename = summary_file
        self.train_log_folder = train_log_folder

    def plot_metric(self, metric_name, **kwargs):
        """
        Plot the training and validation history of a metric.

        This will read out data from the summary file, as well as
        all training log files, and plot them over the epoch.

        Parameters
        ----------
        metric_name : str
            Name of the metric to be plotted over the epoch. This name is what
            was written in the head line of the summary.txt file, except without
            the train_ or val_ prefix.
        kwargs
            Keyword arguments for the plot_history function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The plot.

        """
        summary_data = self.get_summary_data()
        full_train_data = self.get_train_data()
        summary_label = "val_" + metric_name

        if metric_name not in full_train_data.dtype.names:
            raise ValueError(
                "Train log metric name {} unknown, must be one of {}".format(
                    metric_name, self.get_metrics()))
        if summary_label not in summary_data.dtype.names:
            raise ValueError(
                "Summary metric name {} unknown, must be one of {}".format(
                    summary_label, self.get_metrics()))

        if summary_data["Epoch"].shape == (0,):
            # When no lines are present in the summary.txt file.
            val_data = None
        else:
            val_data = [summary_data["Epoch"], summary_data[summary_label]]

        if full_train_data["Batch_float"].shape == (0,):
            # When no lines are present
            raise ValueError("Can not make summary plot: Training log files "
                             "contain no data!")
        else:
            train_data = [full_train_data["Batch_float"],
                          full_train_data[metric_name]]

        # if no validation has been done yet
        if np.all(np.isnan(val_data)[1]):
            val_data = None

        if "y_label" not in kwargs:
            kwargs["y_label"] = metric_name

        fig = plot_history(train_data, val_data, **kwargs)
        return fig

    def plot_lr(self, **kwargs):
        """
        Plot the learning rate over the epochs.

        Parameters
        ----------
        kwargs
            Keyword arguments for the plot_history function.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The plot.

        """
        summary_data = self.get_summary_data()

        epoch = summary_data["Epoch"]
        lr = summary_data["LR"]
        # plot learning rate like val data (connected dots)
        val_data = (epoch, lr)

        if "y_label" not in kwargs:
            kwargs["y_label"] = "Learning rate"
        if "legend" not in kwargs:
            kwargs["legend"] = False

        fig = plot_history(train_data=None, val_data=val_data, **kwargs)
        return fig

    def get_metrics(self):
        """
        Get the name of the metrics from the first line in the summary file.

        This will be the actual name of the metric, i.e. "loss" and not
        "train_loss" or "val_loss".

        Returns
        -------
        all_metrics : List
            A list of the metrics.

        """
        summary_data = self.get_summary_data()
        all_metrics = []
        for keyword in summary_data.dtype.names:
            if keyword == "Epoch" or keyword == "LR":
                continue
            if "train_" in keyword:
                keyword = keyword.split("train_")[-1]
            else:
                keyword = keyword.split("val_")[-1]
            if keyword not in all_metrics:
                all_metrics.append(keyword)
        return all_metrics

    def get_summary_data(self):
        """
        Read out the summary file in the output folder.

        Returns
        -------
        summary_data : ndarray
            Numpy structured array with the column names as datatypes.
            Its shape is the number of lines with data.

        """
        summary_data = np.genfromtxt(self.summary_filename, names=True,
                                     delimiter="|", autostrip=True,
                                     comments="--")
        if summary_data.shape == ():
            # When only one line is present
            summary_data = summary_data.reshape(1,)
        return summary_data

    def get_column_names(self):
        """
        Get the str in the first line in each column.

        Returns
        -------
        tuple : column_names
            The names in the same order as they appear in the summary.txt.

        """
        summary_data = self.get_summary_data()
        column_names = summary_data.dtype.names
        return column_names

    def get_train_data(self):
        """
        Read out all training logfiles in the output folder.

        Read out the data from the summary.txt file, and from all training
        log files in the train_log folder, which is in the same directory
        as the summary.txt file.

        Returns
        -------
        summary_data : numpy.ndarray
            Structured array containing the data from the summary.txt file.
            Its shape is the number of lines with data.

        """
        # list of all files in the train_log folder of this model
        files = os.listdir(self.train_log_folder)
        train_file_data = []
        for file in files:
            if not (file.startswith("log_epoch_") and file.endswith(".txt")):
                continue
            # file is sth like "log_epoch_1_file_2.txt", extract the 1 and 2:
            epoch, file_no = [int(file.split(".")[0].split("_")[i]) for i in [2, 4]]
            file_data = np.genfromtxt(self.train_log_folder + "/" + file,
                                      names=True,
                                      delimiter="|",
                                      autostrip=True,
                                      comments="--")
            train_file_data.append([[epoch, file_no], file_data])

        # sort so that earlier epochs come first
        train_file_data.sort()
        full_train_data = train_file_data[0][1]
        for [epoch, file_no], file_data in train_file_data[1:]:
            full_train_data = np.append(full_train_data, file_data)

        if full_train_data.shape == ():
            # When only one line is present
            full_train_data = full_train_data.reshape(1,)
        return full_train_data

    def get_state(self):
        """
        Get the state of a training.

        For every line in the summary logfile, get a dict with the epoch
        as a float, and is_trained and is_validated bools.

        Returns
        -------
        state_dicts : List
            List of dicts.

        """
        summary_data = self.get_summary_data()
        state_dicts = []
        names = summary_data.dtype.names

        for line in summary_data:
            val_losses, train_losses = {}, {}
            for name in names:
                if name.startswith("val_"):
                    val_losses[name] = line[name]
                elif name.startswith("train_"):
                    train_losses[name] = line[name]
                elif name not in ["Epoch", "LR"]:
                    raise NameError(
                        "Invalid summary file: Invalid column name {}: must be "
                        "either Epoch, LR, or start with val_ or train_".format(name))

            n_nans_val = np.count_nonzero(np.isnan(list(val_losses.values())))
            n_nans_train = np.count_nonzero(np.isnan(list(train_losses.values())))

            if n_nans_val == 0:
                is_val = True
            elif n_nans_val == len(val_losses):
                is_val = False
            else:
                raise ValueError(
                    "Invalid summary file: Expected val losses to be either only "
                    "nans, or no nans (got {})".format(val_losses))

            if n_nans_train == 0:
                is_trained = True
            elif n_nans_train == len(train_losses):
                is_trained = False
            else:
                raise ValueError(
                    "Invalid summary file: Expected train losses to be either only "
                    "nans, or no nans (got {})".format(train_losses))

            line_state = {"epoch": line["Epoch"],
                          "is_trained": is_trained,
                          "is_validated": is_val,}
            state_dicts.append(line_state)

        return state_dicts


def h5_get_number_of_rows(h5_filepath, datasets):
    """
    Gets the total number of rows of of a .h5 file.

    Multiple dataset names can be given as a list to check if they all
    have the same number of rows (axis 0).

    Parameters
    ----------
    h5_filepath : str
        filepath of the .h5 file.
    datasets : list
        The names of datasets in the file to check.

    Returns
    -------
    number_of_rows: int
        number of rows of the .h5 file in the first dataset.

    Raises
    ------
    AssertionError
        If the given datasets do not have the same no of rows.

    """
    with h5py.File(h5_filepath, 'r') as f:
        number_of_rows = [f[dataset].shape[0] for dataset in datasets]
    if not number_of_rows.count(number_of_rows[0]) == len(number_of_rows):
        err_msg = "Datasets do not have the same number of samples " \
                  "in file " + h5_filepath
        for i, dataset in enumerate(datasets):
            err_msg += "\nDataset: {}\tSamples: {}".format(dataset,
                                                           number_of_rows[i])
        raise AssertionError(err_msg)
    return number_of_rows[0]


def use_local_tmpdir(train_files, val_files):
    """
    Copies the val and train files to the local temp folder.

    Returns the new filepaths of the train and val data.
    Speeds up I/O and reduces network load.

    Parameters
    ----------
    train_files : dict
        Dict containing the train file pathes.
    val_files
        Dict containing the val file pathes.

    Returns
    -------
    train_files_ssd : dict
        Train dict with updated SSD /scratch filepaths.
    test_files_ssd : dict
        Val dict with updated SSD /scratch filepaths.

    """
    local_scratch_path = os.environ['TMPDIR']
    train_files_ssd, val_files_ssd = {}, {}

    for input_key in train_files:
        old_pathes = train_files[input_key]
        new_pathes = []
        for f_path in old_pathes:
            # copy to /scratch node-local SSD
            f_path_ssd = local_scratch_path + '/' + os.path.basename(f_path)
            print("Copying", f_path, "\nto", f_path_ssd)
            shutil.copy2(f_path, local_scratch_path)
            new_pathes.append(f_path_ssd)
        train_files_ssd[input_key] = tuple(new_pathes)

    for input_key in val_files:
        old_pathes = val_files[input_key]
        new_pathes = []
        for f_path in old_pathes:
            # copy to /scratch node-local SSD
            f_path_ssd = local_scratch_path + '/' + os.path.basename(f_path)
            print("Copying", f_path, "\nto", f_path_ssd)
            shutil.copy2(f_path, local_scratch_path)
            new_pathes.append(f_path_ssd)
        val_files_ssd[input_key] = tuple(new_pathes)

    print('Finished copying the input data to the local tmpdir folder.')
    return train_files_ssd, val_files_ssd
